use crate::{
    box_vec::BoxVec,
    reckless::{self, with_rounding_up},
    traits::{matrix::Matrix, num::InvMat, Interval},
};
use anyhow::anyhow;
use nalgebra::{ClosedAdd, ClosedMul, ComplexField, DMatrix, Scalar};
use num_traits::{One, Zero};
use std::ops::{Index, IndexMut, Mul, Neg};

#[derive(Clone, Debug)]
pub struct BoxMat<T>(pub DMatrix<T>);

impl<T: Scalar> Matrix for BoxMat<T> {
    type Coeffs = T;

    fn from_value(i: usize, j: usize, v: Self::Coeffs) -> Self {
        Self(DMatrix::from_element(i, j, v))
    }

    fn from_vec(i: usize, j: usize, v: Vec<Self::Coeffs>) -> Self {
        Self(DMatrix::from_vec(i, j, v))
    }

    fn get_unchecked_mut(&mut self, i: usize, j: usize) -> &mut Self::Coeffs {
        unsafe { self.0.get_unchecked_mut((i, j)) }
    }

    fn id(i: usize) -> Self
    where
        Self::Coeffs: num_traits::Zero + num_traits::One,
    {
        Self(DMatrix::identity(i, i))
    }

    fn ncols(&self) -> usize {
        self.0.ncols()
    }

    fn nrows(&self) -> usize {
        self.0.nrows()
    }

    fn zeros(i: usize, j: usize) -> Self
    where
        Self::Coeffs: num_traits::Zero,
    {
        Self(DMatrix::zeros(i, j))
    }
}

impl<T> Index<(usize, usize)> for BoxMat<T> {
    type Output = T;

    fn index(&self, index: (usize, usize)) -> &Self::Output {
        &self.0[index]
    }
}

impl<T> IndexMut<(usize, usize)> for BoxMat<T> {
    fn index_mut(&mut self, index: (usize, usize)) -> &mut Self::Output {
        &mut self.0[index]
    }
}

impl<T> Mul<BoxMat<T>> for BoxMat<T>
where
    T: Scalar + ClosedMul + ClosedAdd + Zero + One,
{
    type Output = Self;

    fn mul(self, rhs: BoxMat<T>) -> Self::Output {
        with_rounding_up(|| Self(self.0 * rhs.0))
    }
}

impl<T> Interval for BoxMat<T>
where
    T: Interval + Scalar + Zero,
{
    fn diam(&self) -> f64 {
        self.0.map(|c| c.diam()).max()
    }

    fn mag(&self) -> f64 {
        self.0.map(|c| c.mag()).max()
    }

    fn mag_lower(&self) -> f64 {
        self.0.map(|c| c.mag_lower()).max()
    }

    fn mut_squash(&mut self) {
        self.0.apply(|c| c.mut_squash());
    }

    fn square(rad: f64) -> Self {
        Self::from_value(1, 1, T::square(rad))
    }

    fn squash(&self) -> Self {
        BoxMat(self.0.map(|c| c.squash()))
    }

    fn union(&self, other: &Self) -> Self {
        debug_assert!(
            self.nrows() == other.nrows() && self.ncols() == other.ncols(),
            "Matrix dimensions do not match"
        );

        let mut res = Self::zeros(self.nrows(), other.ncols());
        for i in 0..self.nrows() {
            for j in 0..self.ncols() {
                res[(i, j)] = self[(i, j)].union(&other[(i, j)])
            }
        }
        res
    }
}

impl<T> BoxMat<T>
where
    T: Scalar + ComplexField,
{
    pub fn try_inverse(self) -> Option<BoxMat<T>> {
        if let Some(v) = self.0.try_inverse() {
            return Some(Self(v));
        }
        None
    }
}

impl<T> Mul<&BoxVec<T>> for &BoxMat<T>
where
    T: Scalar + Zero + One + ClosedMul + ClosedAdd,
{
    type Output = BoxVec<T>;

    fn mul(self, rhs: &BoxVec<T>) -> Self::Output {
        BoxVec(with_rounding_up(|| &self.0 * &rhs.0))
    }
}

impl<T> Mul<&BoxVec<T>> for BoxMat<T>
where
    T: Scalar + Zero + One + ClosedMul + ClosedAdd,
{
    type Output = BoxVec<T>;

    fn mul(self, rhs: &BoxVec<T>) -> Self::Output {
        BoxVec(with_rounding_up(|| self.0 * &rhs.0))
    }
}

impl<T> Mul<BoxVec<T>> for &BoxMat<T>
where
    T: Scalar + Zero + One + ClosedMul + ClosedAdd,
{
    type Output = BoxVec<T>;

    fn mul(self, rhs: BoxVec<T>) -> Self::Output {
        BoxVec(with_rounding_up(|| &self.0 * rhs.0))
    }
}

impl<T> Mul<BoxVec<T>> for BoxMat<T>
where
    T: Scalar + Zero + One + ClosedMul + ClosedAdd,
{
    type Output = BoxVec<T>;

    fn mul(self, rhs: BoxVec<T>) -> Self::Output {
        BoxVec(with_rounding_up(|| self.0 * rhs.0))
    }
}

impl<T> Neg for BoxMat<T>
where
    T: Scalar + Neg<Output = T>,
{
    type Output = BoxMat<T>;

    fn neg(self) -> Self::Output {
        BoxMat(-self.0)
    }
}

impl<T> Neg for &BoxMat<T>
where
    T: Scalar + Neg<Output = T>,
{
    type Output = BoxMat<T>;

    fn neg(self) -> Self::Output {
        BoxMat(-&self.0)
    }
}

impl InvMat for BoxMat<reckless::Cif> {
    fn inv_mat(m: &Self) -> anyhow::Result<Self> {
        let mid_df = BoxMat(m.0.map(|c| c.mid()));
        let inv = with_rounding_up(|| {
            mid_df.try_inverse().ok_or(anyhow!(
                "could not inverse a matrix in floating point arithmetic"
            ))
        })?;

        let mut invdfvec = Vec::new();
        invdfvec.reserve_exact(inv.nrows());
        for c in &inv.0 {
            invdfvec.push(reckless::Cif::new(
                reckless::Rif::new(c.re, c.re),
                reckless::Rif::new(c.im, c.im),
            ))
        }


        let res = BoxMat::from_vec(inv.nrows(), inv.ncols(), invdfvec);
        Ok(res)
    }
}
