use crate::{
    reckless::with_rounding_up,
    traits::{num::RefNeg, vector::Vector, Interval, TaylorModel},
};
use nalgebra::{ClosedAdd, ClosedSub, DVector, Scalar};
use num_traits::Zero;
use std::ops::{Add, AddAssign, Div, DivAssign, Index, IndexMut, Mul, MulAssign, Neg, Sub, SubAssign};

#[derive(Clone, Debug)]
pub struct BoxVec<T>(pub DVector<T>);

impl<T: Scalar + Zero> Vector for BoxVec<T> {
    type T = T;

    fn from_value(i: usize, v: Self::T) -> Self {
        Self(DVector::from_element(i, v))
    }

    fn get_unchecked_mut(&mut self, i: usize) -> &mut Self::T {
        unsafe { self.0.get_unchecked_mut(i) }
    }

    fn len(&self) -> usize {
        self.0.nrows()
    }

    fn zeros(i: usize) -> Self {
        Self(DVector::zeros(i))
    }
}

impl<T> PartialEq for BoxVec<T>
where
    T: PartialEq,
{
    fn eq(&self, other: &Self) -> bool {
        self.0.eq(&other.0)
    }

    fn ne(&self, other: &Self) -> bool {
        self.0.ne(&other.0)
    }
}

impl<T> Index<usize> for BoxVec<T> {
    type Output = T;

    fn index(&self, index: usize) -> &Self::Output {
        &self.0[index]
    }
}

impl<T> IndexMut<usize> for BoxVec<T> {
    fn index_mut(&mut self, index: usize) -> &mut Self::Output {
        &mut self.0[index]
    }
}

impl<T: Scalar> From<Vec<T>> for BoxVec<T> {
    fn from(value: Vec<T>) -> Self {
        Self(DVector::from_vec(value))
    }
}

impl<T> Index<(usize, usize)> for BoxVec<T> {
    type Output = T;

    fn index(&self, index: (usize, usize)) -> &Self::Output {
        debug_assert!(index.1 == 0);
        &self[index.0]
    }
}

impl<T> IndexMut<(usize, usize)> for BoxVec<T> {
    fn index_mut(&mut self, index: (usize, usize)) -> &mut Self::Output {
        debug_assert!(index.1 == 0);
        &mut self[index.0]
    }
}

// Arithmetic operations
//. Addition
impl<T: Scalar + ClosedAdd> Add<BoxVec<T>> for BoxVec<T> {
    type Output = BoxVec<T>;

    fn add(self, rhs: BoxVec<T>) -> Self::Output {
        Self(with_rounding_up(|| self.0 + rhs.0))
    }
}

impl<'a, T: Scalar + ClosedAdd> Add<&'a BoxVec<T>> for BoxVec<T> {
    type Output = BoxVec<T>;

    fn add(self, rhs: &'a BoxVec<T>) -> Self::Output {
        Self(with_rounding_up(|| self.0 + &rhs.0))
    }
}

impl<'a, T: Scalar + ClosedAdd> AddAssign<&'a BoxVec<T>> for BoxVec<T> {
    fn add_assign(&mut self, rhs: &'a BoxVec<T>) {
        with_rounding_up(|| {
            self.0 += &rhs.0;
        });
    }
}

//. Substraction
impl<'a, T: Scalar + ClosedSub> Sub<&'a BoxVec<T>> for BoxVec<T> {
    type Output = BoxVec<T>;

    fn sub(self, rhs: &'a BoxVec<T>) -> Self::Output {
        Self(with_rounding_up(|| self.0 - &rhs.0))
    }
}

impl<'a, T: Scalar + ClosedSub> SubAssign<&'a BoxVec<T>> for BoxVec<T> {
    fn sub_assign(&mut self, rhs: &'a BoxVec<T>) {
        with_rounding_up(|| {
            self.0 -= &rhs.0;
        });
    }
}

impl<T: Scalar + ClosedSub> Sub<BoxVec<T>> for BoxVec<T> {
    type Output = BoxVec<T>;

    fn sub(self, rhs: BoxVec<T>) -> Self::Output {
        Self(with_rounding_up(|| self.0 - rhs.0))
    }
}

//. Multiplication
impl<'a, S, T> Mul<&'a S> for BoxVec<T>
where
    T: MulAssign<&'a S> + Scalar,
    S: Clone,
{
    type Output = BoxVec<T>;

    fn mul(mut self, rhs: &'a S) -> Self::Output {
        with_rounding_up(|| {
            self.0.apply(|c| *c *= rhs);
            self
        })
    }
}

impl<'a, S, T> MulAssign<&'a S> for BoxVec<T>
where
    T: MulAssign<&'a S> + Scalar,
{
    fn mul_assign(&mut self, rhs: &'a S) {
        with_rounding_up(|| {
            self.0.apply(|c| *c *= rhs);
        });
    }
}

impl<'a, 'b, S, T> Mul<&'b S> for &'a BoxVec<T>
where
    T: Scalar + Mul<&'b S, Output = T>,
    S: Scalar + Copy,
{
    type Output = BoxVec<T>;

    fn mul(self, rhs: &'b S) -> Self::Output {
        with_rounding_up(|| BoxVec(self.0.map(|c| c * rhs)))
    }
}

impl<T> TaylorModel for BoxVec<T>
where
    BoxVec<T::Coeffs>: Vector<T = T::Coeffs> + Clone,
    BoxVec<T>: Vector<T = T>,
    T: TaylorModel + Scalar + Zero,
    T::Coeffs: Clone + Scalar,
    T::Domain: Clone,
{
    type Coeffs = BoxVec<T::Coeffs>;
    type Domain = T::Domain;

    // Sloppy
    fn order(&self) -> usize {
        self.0[0].order()
    }

    fn from_vec(dom: &Self::Domain, coeffs: &Vec<Self::Coeffs>) -> Self {
        debug_assert!(!coeffs.is_empty());
        let n = coeffs[0].len();
        debug_assert!(coeffs.iter().all(|c| c.len() == n));
        let mut res: BoxVec<T> = BoxVec::zeros(n);
        for i in 0..res.len() {
            debug_assert!(coeffs.len() <= res[i].order(), "Coefficient vector is too large"); //TODO: improve this
            res[i].set_domain(dom);
            for (j, coeffs_j) in coeffs.iter().enumerate() {
                res[i].set_coeff(coeffs_j[i].clone(), j);
            }
        }
        res
    }

    fn affine(a: &Self::Coeffs, b: &Self::Coeffs, domain: &Self::Domain) -> Self {
        let n = a.len();
        debug_assert!(n == b.len());
        Self(DVector::from_fn(n, |i, _| {
            T::affine(&a[i].clone(), &b[i].clone(), &domain.clone())
        }))
    }

    // Sloppy
    fn domain(&self) -> Self::Domain {
        self[0].domain()
    }

    fn dominter(&mut self, dom: &Self::Domain) {
        self.0.apply(|c| c.dominter(dom));
    }

    fn enclosure(&self, dom: &Self::Domain) -> Self::Coeffs {
        BoxVec(self.0.map(|c| c.enclosure(dom)))
    }

    fn get_coeff_mut(&mut self, _: usize) -> &mut Self::Coeffs {
        // BoxVec(self.0.map(|c| c.get_coeff_mut(i)))
        unimplemented!()
    }

    fn identity(domain: &Self::Domain) -> Self {
        BoxVec(DVector::from_element(1, T::identity(domain)))
    }

    fn set_coeff(&mut self, c: Self::Coeffs, i: usize) {
        self.0.zip_apply(&c.0, |l, r| *l.get_coeff_mut(i) = r);
    }

    fn set_domain(&mut self, dom: &Self::Domain) {
        self.0.apply(|c| c.set_domain(dom));
    }
}

impl<T> Interval for BoxVec<T>
where
    T: Interval + Scalar + Zero,
{
    fn diam(&self) -> f64 {
        self.0.map(|c| c.diam()).max()
    }

    fn mag(&self) -> f64 {
        self.0.map(|c| c.mag()).max()
    }

    fn mag_lower(&self) -> f64 {
        self.0.map(|c| c.mag_lower()).max()
    }

    fn mut_squash(&mut self) {
        self.0.apply(|c| c.mut_squash());
    }

    fn square(rad: f64) -> Self {
        Self::from_value(1, T::square(rad))
    }

    fn squash(&self) -> Self {
        BoxVec(self.0.map(|c| c.squash()))
    }

    fn union(&self, other: &Self) -> Self {
        debug_assert!(self.len() == other.len(), "Vector size do not match");

        let mut res = Self::zeros(self.len());
        for i in 0..self.len() {
            res[i] = self[i].union(&other[i])
        }
        res
    }
}

impl<'a, 'b, T> Sub<&'a BoxVec<T>> for &'b BoxVec<T>
where
    T: Scalar + ClosedSub,
{
    type Output = BoxVec<T>;

    fn sub(self, rhs: &'a BoxVec<T>) -> Self::Output {
        BoxVec(with_rounding_up(|| &self.0 - &rhs.0))
    }
}

impl<'a, 'b, T> Add<&'a BoxVec<T>> for &'b BoxVec<T>
where
    T: Scalar + ClosedAdd,
{
    type Output = BoxVec<T>;

    fn add(self, rhs: &'a BoxVec<T>) -> Self::Output {
        BoxVec(with_rounding_up(|| &self.0 + &rhs.0))
    }
}

impl<T> Neg for BoxVec<T>
where
    T: Scalar + Neg<Output = T>,
{
    type Output = BoxVec<T>;

    fn neg(self) -> Self::Output {
        BoxVec(-self.0)
    }
}

impl<T> Div<T> for BoxVec<T>
where
    for<'a> T: DivAssign<&'a T>,
{
    type Output = BoxVec<T>;

    fn div(mut self, rhs: T) -> Self::Output {
        with_rounding_up(|| {
            self.0.apply(|c| *c /= &rhs);
            self
        })
    }
}

impl<T> Sub<BoxVec<T>> for &BoxVec<T>
where
    BoxVec<T>: RefNeg,
    for<'r> BoxVec<T>: SubAssign<&'r BoxVec<T>>,
{
    type Output = BoxVec<T>;

    fn sub(self, mut rhs: BoxVec<T>) -> Self::Output {
        rhs -= self;
        rhs.ref_neg();
        rhs
    }
}

impl<T> RefNeg for BoxVec<T>
where
    T: RefNeg,
{
    fn ref_neg(&mut self) {
        self.0.apply(|c| c.ref_neg());
    }
}
