// Copyright 2024 Pierre Lairez
// SPDX-License-Identifier: GPL-3.0-or-later

// COMMAND LINE INTERFACE

// OPTIONS

use crate::{
    circuit::{self, Circuit},
    flint::Fmpq,
    prog::Program,
    projective::ParametricSystem,
};
use anyhow::bail;
use itertools::{izip, Itertools};
use nalgebra::{one, Complex};
use num_traits::{One, Zero};
use std::{collections::HashMap, hash::Hash, io::Read};

#[derive(Clone, Copy, clap::ValueEnum)]
pub enum Arithmetic {
    Reckless,
    Arb,
    ArbReckless,
}

#[derive(clap::Parser)]
#[command(author, version, about)]
pub struct Configuration {
    /// puts the equations into Horner form before processing
    #[arg(long)]
    pub horner: bool,

    /// enable homogenization of equations, does nothing if the input is already homogeneous
    #[arg(long)]
    pub homogenize: bool,

    /// input in legacy format
    #[arg(long, hide = true)]
    pub legacy: bool,

    /// differentiation mode for circuits
    #[arg(long, default_value = "forward")]
    pub diff: circuit::DifferentiationMode,

    /// Number of threads. Default to the number of available CPUs, or 1 if the
    /// program fails to determine this number.
    #[arg(long)]
    pub jobs: Option<usize>,

    /// Input file in JSON syntax. If none given, read standard input.
    pub path: Option<std::path::PathBuf>,

    #[arg(long, hide = true)]
    pub markdown_help: bool,

    #[arg(long, default_value = "reckless")]
    pub arithmetic: Arithmetic,
}

impl Default for Configuration {
    fn default() -> Self {
        Configuration {
            horner: false,
            homogenize: false,
            legacy: false,
            diff: circuit::DifferentiationMode::Forward,
            jobs: None,
            path: None,
            markdown_help: false,
            arithmetic: Arithmetic::Reckless,
        }
    }
}

// JSON INPUT

#[derive(Clone)]
pub struct Problem {
    pub system: Vec<Circuit<Complex<Fmpq>>>,
    pub variables: Vec<Circuit<Complex<Fmpq>>>,
    pub parameters: Vec<Circuit<Complex<Fmpq>>>,
    pub path: Vec<Vec<Complex<Fmpq>>>,
    pub fiber: Vec<Vec<Complex<Fmpq>>>,
}

#[derive(serde::Deserialize, Clone)]
pub struct _Problem {
    pub system: Vec<Circuit<Complex<Fmpq>>>,
    pub variables: Vec<Circuit<Complex<Fmpq>>>,
    pub parameters: Vec<Circuit<Complex<Fmpq>>>,
    pub path: Vec<Vec<Circuit<Complex<Fmpq>>>>,
    pub fiber: Vec<Vec<Circuit<Complex<Fmpq>>>>,
}

impl TryFrom<_Problem> for Problem {
    type Error = anyhow::Error;

    fn try_from(value: _Problem) -> Result<Self, Self::Error> {
        let try_path: anyhow::Result<Vec<Vec<Complex<Fmpq>>>> = value
            .path
            .into_iter()
            .map(|v| {
                v.into_iter()
                    .map(|circ| match circ._get_scalar() {
                        None => bail!("Path points must be scalar"),
                        Some(x) => Ok(x),
                    })
                    .collect::<anyhow::Result<Vec<Complex<Fmpq>>>>()
            })
            .collect::<anyhow::Result<Vec<Vec<Complex<Fmpq>>>>>();

        let path = match try_path {
            Err(e) => return Err(e),
            Ok(p) => p,
        };

        let try_fiber: anyhow::Result<Vec<Vec<Complex<Fmpq>>>> = value
            .fiber
            .into_iter()
            .map(|v| {
                v.into_iter()
                    .map(|circ| match circ._get_scalar() {
                        None => bail!("Fiber points must be scalar"),
                        Some(x) => Ok(x),
                    })
                    .collect::<anyhow::Result<Vec<Complex<Fmpq>>>>()
            })
            .collect::<anyhow::Result<Vec<Vec<Complex<Fmpq>>>>>();

        let fiber = match try_fiber {
            Err(e) => return Err(e),
            Ok(f) => f,
        };

        Ok(Problem {
            system: value.system,
            variables: value.variables,
            parameters: value.parameters,
            path,
            fiber,
        })
    }
}

// #[derive(serde::Deserialize, Clone)]
// struct ProblemOldFormat {
//     vars: Vec<Circuit<Complex<Fmpq>>>,
//     parameter: Circuit<Complex<Fmpq>>,
//     sys: Vec<Circuit<Complex<Fmpq>>>,
//     initial_sols: Vec<Vec<Complex<Fmpq>>>,
// }

// impl From<ProblemOldFormat> for Problem {
//     fn from(pb: ProblemOldFormat) -> Self {
//         Problem {
//             variables: pb.vars,
//             parameters: vec![pb.parameter],
//             system: pb.sys,
//             path: vec![vec![Cif::zero()], vec![Cif::one()]],
//             fiber: pb.initial_sols,
//         }
//     }
// }

impl Problem {
    pub fn new_from_string(str: &str, conf: &Configuration) -> Result<Problem, anyhow::Error> {
        let pb: _Problem = if conf.legacy {
            panic!(); //TODO
                      // let pb_legacy_format: ProblemOldFormat = serde_json::from_str(str)?;
                      // pb = pb_legacy_format.into();
        } else {
            serde_json::from_str(str)?
        };
        pb.try_into()
    }

    pub fn new_from_input(conf: &Configuration) -> Result<Problem, anyhow::Error> {
        let mut input = String::new();
        if let Some(path) = &conf.path {
            let mut fs = std::fs::File::open(path)?;
            fs.read_to_string(&mut input).unwrap();
        } else {
            std::io::stdin().read_to_string(&mut input)?;
        }

        Self::new_from_string(&input, conf)
    }

    pub fn build_parametric_system(
        &mut self,
        conf: &Configuration,
    ) -> Result<ParametricSystem<Program<Complex<Fmpq>>>, std::io::Error> {
        let dim = self.system.len();
        let nparams = self.parameters.len();
        assert!(dim == self.variables.len());

        let mut allvars = Vec::new();
        allvars.extend_from_slice(&self.parameters);
        allvars.extend_from_slice(&self.variables);

        if conf.horner {
            for eq in &mut self.system {
                *eq = eq.horner(&allvars);
            }
        }

        let jac = Circuit::jacobian_matrix(&self.system, &allvars, conf.diff);
        let f0 = Program::compile(&self.system, &allvars);

        let jacp = jac.columns(0, self.parameters.len());
        let fdot0 = Program::compile(&jacp.iter().cloned().collect(), &allvars);

        let jacv = jac.columns(self.parameters.len(), self.variables.len());
        let df0 = Program::compile(&jacv.iter().cloned().collect(), &allvars);

        let mut f = vec![f0];
        let mut df = vec![df0];
        let mut fdot = vec![fdot0];

        let mut chart_switch_threshold = f64::INFINITY;

        if conf.homogenize {
            let ctx = self.variables[0].context();
            chart_switch_threshold = 2.0;
            let homogenization_var = ctx.var("__homogenization_variable__");

            let mut homogenized_equations = Vec::new();
            for eq in &mut self.system {
                let (_degree, heq) = eq.homogenize(&self.variables, &homogenization_var);
                homogenized_equations.push(heq);
            }

            for i in 0..dim {
                let mut dict = HashMap::new();
                dict.insert(self.variables[i].clone(), ctx.scalar(Complex::one()));
                let sys_in_chart: Vec<_> = homogenized_equations.iter().map(|c| c.subs_with_memo(&mut dict)).collect();

                let mut allvars = Vec::new();
                allvars.extend_from_slice(&self.parameters);
                allvars.push(homogenization_var.clone());
                allvars.extend_from_slice(&self.variables[..i]);
                allvars.extend_from_slice(&self.variables[i + 1..]);

                let jac = Circuit::jacobian_matrix(&sys_in_chart, &allvars, conf.diff);
                f.push(Program::compile(&sys_in_chart, &allvars));

                let jacp = jac.columns(0, self.parameters.len());
                fdot.push(Program::compile(&jacp.iter().cloned().collect(), &allvars));

                let jacv = jac.columns(self.parameters.len(), self.variables.len());
                df.push(Program::compile(&jacv.iter().cloned().collect(), &allvars));
            }
        }

        Ok(ParametricSystem {
            nparams,
            dim,
            f,
            df,
            fdot,
            chart_switch_threshold,
        })
    }

    // Transfoms a problem of following the roots of some system along a path into following the roots of multiple systems consecutively along [0, 1] each
    pub fn precompose_path(&self) -> anyhow::Result<Vec<Problem>> {
        let mut pt: HashMap<Circuit<Complex<Fmpq>>, Circuit<Complex<Fmpq>>> = HashMap::default();
        let ctx = self.variables[0].context();
        let parameter_variable = ctx.var("__parameter__");
        let mut res = vec![];
        for (i, (p, q)) in self.path.iter().tuple_windows().enumerate() {
            // Build linear interpolation between p and q
            for (pi, qi, v) in izip!(p, q, &self.parameters) {
                pt.insert(
                    v.clone(),
                    (ctx.scalar(Complex::<Fmpq>::one()) - &parameter_variable) * ctx.scalar(pi.clone())
                        + ctx.scalar(qi.clone()) * &parameter_variable,
                );
            }

            // Precompose the system with the linear interpolation
            let system = self
                .system
                .iter()
                .map(|eq| match eq.compose(&pt) {
                    None => bail!("composition failed"),
                    Some(composed_eq) => Ok(composed_eq),
                })
                .collect::<anyhow::Result<_>>()?;

            let fiber = if i == 0 { self.fiber.clone() } else { vec![] };

            res.push(Problem {
                system: system,
                variables: self.variables.clone(),
                path: vec![],
                parameters: vec![parameter_variable.clone()],
                fiber: fiber,
            })
        }
        Ok(res)
    }
}

#[cfg(test)]
mod precompose_path_test {
    use std::path::PathBuf;

    use super::{Configuration, Problem};

    #[test]
    fn test() {

        let path = PathBuf::from("data/2-2_2.json");
        let conf = Configuration {
            path: Some(path),
            ..Default::default()
        };

        let input_problem = Problem::new_from_input(&conf).unwrap();
        println!("{:?}", input_problem.system[0]);
        println!("###################");
        let problems = input_problem.precompose_path().unwrap();
        problems.iter().for_each(|problem| println!("{:?}", problem.system[0]));
    }
}
