use super::{Arb, Fmpq};
use crate::{
    precision_ctx::get_prec,
    reckless::Cif,
    traits::{
        num::{MulWithRop, RefNeg},
        set::Set,
        Interval,
    },
};
use flint_bindings::*;
use nalgebra::Complex;
use num_traits::{One, Zero};
use std::{
    fmt::{Debug, Display},
    mem::MaybeUninit,
    ops::{Add, AddAssign, Div, DivAssign, Mul, MulAssign, Neg},
};

pub struct Acb(pub acb_struct);

// Basics
impl Default for Acb {
    fn default() -> Self {
        unsafe {
            let mut default = MaybeUninit::uninit();
            acb_init(default.as_mut_ptr());
            Self(default.assume_init())
        }
    }
}

impl Clone for Acb {
    fn clone(&self) -> Self {
        let mut res = Self::default();
        unsafe {
            acb_set(&mut res.0, &self.0);
        }
        res
    }
}

impl Drop for Acb {
    fn drop(&mut self) {
        unsafe {
            acb_clear(&mut self.0);
        }
    }
}

impl Debug for Acb {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{:?} + I*{:?}", self.re(), self.im())
    }
}

impl Display for Acb {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{:?}", self)
    }
}

// Specific impls
impl Acb {
    pub fn from_re_im(re: &Arb, im: &Arb) -> Acb {
        Acb(acb_struct {
            real: re.0.clone(),
            imag: im.0.clone(),
        })
    }

    pub fn try_get_included_reckless_interval(&self) -> anyhow::Result<Cif> {
        Ok(Cif::new(
            self.re().try_get_included_reckless_interval()?,
            self.im().try_get_included_reckless_interval()?,
        ))
    }

    pub fn get_covering_reckless_interval(&self) -> Cif {
        Cif::new(
            self.re().get_covering_reckless_interval(),
            self.im().get_covering_reckless_interval(),
        )
    }

    pub fn re(&self) -> &Arb {
        (&self.0.real).into()
    }

    pub fn im(&self) -> &Arb {
        (&self.0.imag).into()
    }
}

// Special values and conversions

impl One for Acb {
    fn one() -> Self {
        let mut one = Self::default();
        one.set_one();
        one
    }

    fn set_one(&mut self) {
        unsafe {
            acb_one(&mut self.0);
        }
    }

    fn is_one(&self) -> bool
    where
        Self: PartialEq,
    {
        unsafe { acb_is_one(&self.0) != 0 }
    }
}

impl Zero for Acb {
    fn zero() -> Self {
        let mut zero = Self::default();
        zero.set_zero();
        zero
    }

    fn set_zero(&mut self) {
        unsafe {
            acb_zero(&mut self.0);
        }
    }

    fn is_zero(&self) -> bool
    where
        Self: PartialEq,
    {
        unsafe { acb_is_zero(&self.0) != 0 }
    }
}

impl From<&acb_struct> for &Acb {
    fn from(value: &acb_struct) -> Self {
        unsafe { &*(value as *const acb_struct as *const Acb) }
    }
}

impl From<&mut acb_struct> for &mut Acb {
    fn from(value: &mut acb_struct) -> Self {
        unsafe { &mut *(value as *mut acb_struct as *mut Acb) }
    }
}

impl From<Arb> for Acb {
    fn from(value: Arb) -> Self {
        //TODO: make this work
        // let mut into = Self::default();
        // into.0.real = value.0;
        // into
        (&value).into()
    }
}

impl From<&Arb> for Acb {
    fn from(value: &Arb) -> Self {
        let mut into = Self::default();
        unsafe {
            acb_set_arb(&mut into.0, &value.0);
        }
        into
    }
}

impl From<i64> for Acb {
    fn from(value: i64) -> Self {
        let mut into = Self::default();
        unsafe {
            acb_set_si(&mut into.0, value);
        }
        into
    }
}

impl From<Complex<Fmpq>> for Acb {
    fn from(value: Complex<Fmpq>) -> Self {
        (&value).into()
    }
}

impl From<&Complex<Fmpq>> for Acb {
    fn from(value: &Complex<Fmpq>) -> Self {
        let mut res = Acb::default();
        res.set(value);
        res
    }
}

impl Set<&Complex<Fmpq>> for Acb {
    fn set(&mut self, other: &Complex<Fmpq>) {
        <&mut Arb>::from(&mut self.0.real).set(&other.re);
        <&mut Arb>::from(&mut self.0.imag).set(&other.im);
    }
}

// Comparisons
impl PartialEq for Acb {
    fn eq(&self, other: &Self) -> bool {
        unsafe { acb_contains(&self.0, &other.0) != 0 && acb_contains(&other.0, &self.0) != 0 }
    }
}

impl Eq for Acb {}

// Arithmetic
impl AddAssign<&Acb> for Acb {
    fn add_assign(&mut self, rhs: &Acb) {
        unsafe {
            acb_add(&mut self.0, &self.0, &rhs.0, get_prec());
        }
    }
}

impl AddAssign<Acb> for Acb {
    fn add_assign(&mut self, rhs: Acb) {
        *self += &rhs;
    }
}

impl Add<Acb> for Acb {
    type Output = Self;

    fn add(self, rhs: Acb) -> Self::Output {
        self + &rhs
    }
}

impl Add<&Acb> for Acb {
    type Output = Acb;

    fn add(mut self, rhs: &Acb) -> Self::Output {
        self += rhs;
        self
    }
}

impl Add<Acb> for &Acb {
    type Output = Acb;

    fn add(self, rhs: Acb) -> Self::Output {
        rhs + self
    }
}

impl MulWithRop<Acb> for Acb {
    type Output = Acb;

    fn mul_with_rop(&self, rhs: &Acb, res: &mut Self::Output) {
        unsafe {
            acb_mul(&mut res.0, &self.0, &rhs.0, get_prec());
        }
    }
}

impl MulAssign<&Acb> for Acb {
    fn mul_assign(&mut self, rhs: &Acb) {
        unsafe {
            acb_mul(&mut self.0, &self.0, &rhs.0, get_prec());
        }
    }
}

impl Mul<&Acb> for Acb {
    type Output = Acb;

    fn mul(mut self, rhs: &Acb) -> Self::Output {
        self *= rhs;
        self
    }
}

impl Mul<Acb> for Acb {
    type Output = Self;

    fn mul(self, rhs: Acb) -> Self::Output {
        self * &rhs
    }
}

impl Mul<&Acb> for &Acb {
    type Output = Acb;

    fn mul(self, rhs: &Acb) -> Self::Output {
        let mut res = Acb::default();
        self.mul_with_rop(rhs, &mut res);
        res
    }
}

impl MulAssign<&Arb> for Acb {
    fn mul_assign(&mut self, rhs: &Arb) {
        unsafe {
            acb_mul_arb(&mut self.0, &self.0, &rhs.0, get_prec());
        }
    }
}

impl Mul<&Arb> for Acb {
    type Output = Acb;

    fn mul(mut self, rhs: &Arb) -> Self::Output {
        self *= rhs;
        self
    }
}

impl Neg for Acb {
    type Output = Self;

    fn neg(mut self) -> Self::Output {
        self.ref_neg();
        self
    }
}

impl RefNeg for Acb {
    fn ref_neg(&mut self) {
        unsafe {
            acb_neg(&mut self.0, &self.0);
        }
    }
}

impl Div<&Acb> for Acb {
    type Output = Self;

    fn div(mut self, rhs: &Acb) -> Self::Output {
        self /= rhs;
        self
    }
}

impl DivAssign<&Acb> for Acb {
    fn div_assign(&mut self, rhs: &Acb) {
        unsafe {
            acb_div(&mut self.0, &self.0, &rhs.0, get_prec());
        }
    }
}

// Interval
impl Interval for Acb {
    fn union(&self, other: &Self) -> Self {
        let mut res = Acb::default();
        unsafe {
            acb_union(&mut res.0, &self.0, &other.0, get_prec());
        }
        res
    }

    fn diam(&self) -> f64 {
        self.re().diam().max(self.im().diam())
    }

    // acb_get_mag returns an upper bound for self for the euclidean norm, whereas we would like an upper bound for the real max norm on ℂ ≃ ℝ².
    fn mag(&self) -> f64 {
        self.re().mag().max(self.im().mag())
    }

    fn mag_lower(&self) -> f64 {
        self.re().mag_lower().max(self.im().mag_lower())
    }

    fn squash(&self) -> Self {
        let mut squashed = Acb::default();
        unsafe { acb_get_mid(&mut squashed.0, &self.0) }
        squashed
    }

    fn mut_squash(&mut self) {
        unsafe { acb_get_mid(&mut self.0, &self.0) }
    }

    fn square(rad: f64) -> Self {
        Acb::from_re_im(&Arb::square(rad), &Arb::square(rad))
    }
}
