use super::{acb_vec::AcbVecIterMut, Acb, AcbVec};
use crate::{
    precision_ctx::get_prec,
    traits::{matrix::Matrix, num::InvMat, vector::Vector, Interval},
};
use anyhow::bail;
use flint_bindings::*;
use num_traits::One;
use std::{
    fmt::Debug,
    marker::PhantomData,
    mem::MaybeUninit,
    ops::{Index, IndexMut, Mul},
    ptr::NonNull,
};

pub struct AcbMat(pub acb_mat_struct);

// Basics
impl Clone for AcbMat {
    fn clone(&self) -> Self {
        let mut clone = Self::new(self.nrows(), self.ncols());
        unsafe {
            acb_mat_set(&mut clone.0, &self.0);
        }
        clone
    }
}

impl Drop for AcbMat {
    fn drop(&mut self) {
        unsafe {
            acb_mat_clear(&mut self.0);
        }
    }
}

impl AcbMat {
    pub fn new(nrows: usize, ncols: usize) -> Self {
        unsafe {
            let mut new = MaybeUninit::uninit();
            acb_mat_init(new.as_mut_ptr(), nrows.try_into().unwrap(), ncols.try_into().unwrap());
            Self(new.assume_init())
        }
    }
}

impl Debug for AcbMat {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        writeln!(f, "AcbMat")?;
        for i in 0..self.nrows() {
            write!(f, "[")?;
            for j in 0..(self.ncols() - 1) {
                write!(f, "{:?}, ", self[(i, j)])?;
            }
            write!(f, "{:?}]", self[(i, self.ncols() - 1)])?;
            if i != self.nrows() - 1 {
                writeln!(f)?;
            }
        }
        Ok(())
    }
}

// Indexing and iterating
impl Index<(usize, usize)> for AcbMat {
    type Output = Acb;

    fn index(&self, index: (usize, usize)) -> &Self::Output {
        debug_assert!(index.0 < self.nrows() && index.1 < self.ncols(), "Index out of bounds");
        unsafe {
            (self
                .0
                .entries
                .add(index.0 * usize::try_from(self.0.stride).unwrap() + index.1))
            .as_ref()
            .unwrap()
            .into()
        }
    }
}

impl IndexMut<(usize, usize)> for AcbMat {
    fn index_mut(&mut self, index: (usize, usize)) -> &mut Self::Output {
        debug_assert!(index.0 < self.nrows() && index.1 < self.ncols(), "Index out of bounds");
        unsafe {
            (self
                .0
                .entries
                .add(index.0 * usize::try_from(self.0.stride).unwrap() + index.1))
            .as_mut()
            .unwrap()
            .into()
        }
    }
}

pub struct AcbMatIter<'a> {
    mat: &'a AcbMat,
    i: usize,
    j: usize,
    nrows: usize,
    ncols: usize,
}

pub struct AcbMatIterMut<'a> {
    mat: &'a AcbMat,
    i: usize,
    nrows: usize,
    row: AcbVecIterMut<'a>,
}

// We iterate left to right then top to bottom.
impl<'a> Iterator for AcbMatIter<'a> {
    type Item = &'a Acb;

    fn next(&mut self) -> Option<Self::Item> {
        let i = self.i;
        let j = self.j;
        let mut res = None;
        if i != self.nrows {
            res = Some(&self.mat[(i, j)]);
            if j == self.ncols - 1 {
                self.j = 0;
                self.i += 1;
            } else {
                self.j += 1;
            }
        }
        res
    }
}

impl<'a> Iterator for AcbMatIterMut<'a> {
    type Item = &'a mut Acb;

    fn next(&mut self) -> Option<Self::Item> {
        let i = self.i;
        match self.row.next() {
            Some(v) => Some(v),
            None => match i + 1 == self.nrows {
                true => None,
                false => {
                    self.i += 1;
                    self.row = AcbVecIterMut {
                        phantom_lifetime: PhantomData,
                        ptr: NonNull::from(&self.mat[(self.i, 0)]),
                        index: 0,
                        len: self.row.len,
                    };
                    self.row.next()
                }
            },
        }
    }
}

impl<'a> IntoIterator for &'a AcbMat {
    type Item = &'a Acb;
    type IntoIter = AcbMatIter<'a>;

    fn into_iter(self) -> Self::IntoIter {
        AcbMatIter {
            mat: self,
            i: 0,
            j: 0,
            nrows: self.nrows(),
            ncols: self.ncols(),
        }
    }
}

impl<'a> IntoIterator for &'a mut AcbMat {
    type Item = &'a mut Acb;
    type IntoIter = AcbMatIterMut<'a>;

    fn into_iter(self) -> Self::IntoIter {
        AcbMatIterMut {
            mat: self,
            i: 0,
            nrows: self.nrows(),
            row: AcbVecIterMut {
                phantom_lifetime: PhantomData,
                ptr: NonNull::from(&self[(0, 0)]),
                index: 0,
                len: self.ncols(),
            },
        }
    }
}

// Arithmetic
impl Mul<AcbVec> for &AcbMat {
    type Output = AcbVec;

    fn mul(self, mut rhs: AcbVec) -> Self::Output {
        if self.ncols() == self.nrows() {
            debug_assert!(self.ncols() == rhs.len(), "Matrix vector product dimension error");
            unsafe {
                acb_mat_mul(&mut rhs.0, &self.0, &rhs.0, get_prec());
            }
            rhs
        } else {
            self * &rhs
        }
    }
}

impl Mul<AcbVec> for AcbMat {
    type Output = AcbVec;

    fn mul(self, rhs: AcbVec) -> Self::Output {
        &self * rhs
    }
}

impl Mul<&AcbVec> for &AcbMat {
    type Output = AcbVec;

    fn mul(self, rhs: &AcbVec) -> Self::Output {
        debug_assert!(self.ncols() == rhs.len(), "Matrix vector product dimension error");
        let mut res = AcbVec::new(self.nrows());
        unsafe {
            acb_mat_mul(&mut res.0, &self.0, &rhs.0, get_prec());
        }
        res
    }
}

impl Mul<&AcbVec> for AcbMat {
    type Output = AcbVec;

    fn mul(self, rhs: &AcbVec) -> Self::Output {
        &self * rhs
    }
}

impl InvMat for AcbMat {
    fn inv_mat(m: &Self) -> anyhow::Result<Self> {
        debug_assert!(m.ncols() == m.nrows(), "Cannot invert non-square matrix");
        let mut res = Self::new(m.nrows(), m.ncols());
        unsafe {
            match acb_mat_inv(&mut res.0, &m.0, get_prec()) {
                0 => bail!("could not inverse a matrix in floating point arithmetic"),
                _ => Ok(res),
            }
        }
    }
}

// Matrix
impl Matrix for AcbMat {
    type Coeffs = Acb;

    fn nrows(&self) -> usize {
        self.0.r.try_into().unwrap()
    }

    fn ncols(&self) -> usize {
        self.0.c.try_into().unwrap()
    }

    fn zeros(i: usize, j: usize) -> Self
    where
        Self::Coeffs: num_traits::Zero,
    {
        Self::new(i, j)
    }

    fn id(n: usize) -> Self
    where
        Self::Coeffs: num_traits::Zero + num_traits::One,
    {
        let mut id = Self::new(n, n);
        for i in 0..n {
            id[(i, i)] = Acb::one();
        }
        id
    }

    fn from_value(m: usize, n: usize, v: Self::Coeffs) -> Self {
        let mut res = Self::new(m, n);
        for i in 0..m {
            for j in 0..n {
                res[(i, j)] = v.clone();
            }
        }
        res
    }

    fn from_vec(nrows: usize, ncols: usize, v: Vec<Self::Coeffs>) -> Self {
        let mut res = Self::new(nrows, ncols);
        let mut cpt = 0;
        for j in 0..ncols {
            for i in 0..nrows {
                if cpt == v.len() {
                    break;
                }
                res[(i, j)] = v[cpt].clone();
                cpt += 1;
            }
        }
        res
    }

    fn get_unchecked_mut(&mut self, i: usize, j: usize) -> &mut Self::Coeffs {
        &mut self[(i, j)]
    }
}

// Interval
impl Interval for AcbMat {
    fn diam(&self) -> f64 {
        self.into_iter()
            .map(|c| c.diam())
            .max_by(|x, y| x.total_cmp(y))
            .unwrap_or(0.0)
    }

    fn mag(&self) -> f64 {
        unimplemented!()
    }

    fn mag_lower(&self) -> f64 {
        unimplemented!()
    }

    fn mut_squash(&mut self) {
        for x in self {
            (*x).mut_squash();
        }
    }

    fn square(_rad: f64) -> Self {
        unimplemented!()
    }

    fn squash(&self) -> Self {
        let mut res = AcbMat::new(self.nrows(), self.ncols());
        for (res_x, self_x) in (&mut res).into_iter().zip(self.into_iter()) {
            *res_x = self_x.squash();
        }
        res
    }

    fn union(&self, _other: &Self) -> Self {
        todo!()
    }
}
