use super::{Acb, Arb};
use crate::{
    box_vec::BoxVec,
    precision_ctx::get_prec,
    taylor_model::Tm,
    traits::{interval::RealInterval, vector::Vector, Interval, TaylorModel},
};
use core::f64;
use flint_bindings::*;
use num_traits::Zero;
use std::{
    fmt::Debug,
    marker::PhantomData,
    mem::MaybeUninit,
    ops::{Add, AddAssign, Div, Index, IndexMut, Mul, Neg, Sub, SubAssign},
    ptr::NonNull,
};

pub struct AcbVec(pub acb_mat_struct);

// Basics
impl AcbVec {
    pub fn new(nrows: usize) -> Self {
        unsafe {
            let mut _new = MaybeUninit::uninit();
            acb_mat_init(_new.as_mut_ptr(), nrows.try_into().unwrap(), 1);
            Self(_new.assume_init())
        }
    }
}

impl Clone for AcbVec {
    fn clone(&self) -> Self {
        let mut clone = AcbVec::new(self.len());
        unsafe {
            acb_mat_set(&mut clone.0, &self.0);
        }
        clone
    }
}

impl Drop for AcbVec {
    fn drop(&mut self) {
        unsafe {
            acb_mat_clear(&mut self.0);
        }
    }
}

impl Debug for AcbVec {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        self.into_iter()
            .enumerate()
            .map(|(i, self_i)| {
                if i == 0 {
                    write!(f, "AcbVec ({:?}", self_i)
                } else {
                    write!(f, ", {:?}", self_i)
                }
            })
            .collect::<std::fmt::Result>()?;
        write!(f, ")")
    }
}

// Indexing and iterating
impl Index<usize> for AcbVec {
    type Output = Acb;

    fn index(&self, index: usize) -> &Self::Output {
        debug_assert!(index < self.len(), "Index out of bounds");
        unsafe { (self.0.entries.add(index)).as_ref().unwrap().into() }
    }
}

impl IndexMut<usize> for AcbVec {
    fn index_mut(&mut self, index: usize) -> &mut Self::Output {
        debug_assert!(index < self.len(), "Index out of bounds");
        unsafe { (self.0.entries.add(index)).as_mut().unwrap().into() }
    }
}

pub struct AcbVecIter<'a> {
    pub phantom_lifetime: PhantomData<&'a ()>,
    pub ptr: NonNull<Acb>,
    pub index: usize,
    pub len: usize,
}

pub struct AcbVecIterMut<'a> {
    pub phantom_lifetime: PhantomData<&'a ()>,
    pub ptr: NonNull<Acb>,
    pub index: usize,
    pub len: usize,
}

impl<'a> Iterator for AcbVecIter<'a> {
    type Item = &'a Acb;

    fn next(&mut self) -> Option<Self::Item> {
        if self.index == self.len {
            None
        } else {
            self.index += 1;
            unsafe {
                let old = self.ptr;
                if self.index != self.len {
                    self.ptr = self.ptr.add(1);
                }
                Some(old.as_ref())
            }
        }
    }

    fn size_hint(&self) -> (usize, Option<usize>) {
        (self.len - self.index, Some(self.len - self.index))
    }
}

impl<'a> Iterator for AcbVecIterMut<'a> {
    type Item = &'a mut Acb;

    fn next(&mut self) -> Option<Self::Item> {
        if self.index == self.len {
            None
        } else {
            self.index += 1;
            unsafe {
                let mut old = self.ptr;
                if self.index != self.len {
                    self.ptr = self.ptr.add(1);
                }
                Some(old.as_mut())
            }
        }
    }

    fn size_hint(&self) -> (usize, Option<usize>) {
        (self.len - self.index, Some(self.len - self.index))
    }
}

impl<'a> IntoIterator for &'a AcbVec {
    type Item = &'a Acb;
    type IntoIter = AcbVecIter<'a>;

    fn into_iter(self) -> Self::IntoIter {
        AcbVecIter {
            phantom_lifetime: PhantomData,
            ptr: NonNull::from(&self[0]).cast(),
            index: 0,
            len: self.len(),
        }
    }
}

impl<'a> IntoIterator for &'a mut AcbVec {
    type Item = &'a mut Acb;
    type IntoIter = AcbVecIterMut<'a>;

    fn into_iter(self) -> Self::IntoIter {
        AcbVecIterMut {
            phantom_lifetime: PhantomData,
            ptr: NonNull::from(&self[0]).cast(),
            index: 0,
            len: self.len(),
        }
    }
}

impl<'a> FromIterator<Acb> for AcbVec {
    fn from_iter<T: IntoIterator<Item = Acb>>(iter: T) -> Self {
        let iter = iter.into_iter();
        match iter.size_hint() {
            (low, Some(upp)) => {
                if low == upp {
                    let mut res = AcbVec::new(upp);
                    for (i, x) in iter.into_iter().enumerate() {
                        res[i] = x;
                    }
                    return res;
                } else {
                    todo!()
                }
            }
            _ => todo!(),
        }
    }
}

// Special values and conversions
impl From<Vec<Acb>> for AcbVec {
    fn from(value: Vec<Acb>) -> Self {
        value.into_iter().collect()
    }
}

impl<const N: usize> From<&AcbVec> for BoxVec<Tm<N, Arb, Acb>> {
    fn from(value: &AcbVec) -> Self {
        let mut res = BoxVec::zeros(value.len());
        for (res_i, self_i) in (&mut res.0).into_iter().zip(value.into_iter()) {
            *res_i = Tm::affine(&Acb::zero(), self_i, &Arb::entire());
        }
        res
    }
}

// Comparisons
impl PartialEq for AcbVec {
    fn eq(&self, other: &Self) -> bool {
        self.into_iter().zip(other.into_iter()).all(|(x, y)| x == y)
    }
}

// Arithmetic
// +
impl AddAssign<&AcbVec> for AcbVec {
    fn add_assign(&mut self, rhs: &AcbVec) {
        debug_assert!(self.len() == rhs.len(), "Lenghts do not match");
        unsafe {
            acb_mat_add(&mut self.0, &self.0, &rhs.0, get_prec());
        }
    }
}

impl Add<AcbVec> for AcbVec {
    type Output = AcbVec;

    fn add(mut self, rhs: AcbVec) -> Self::Output {
        self += &rhs;
        self
    }
}

impl Add<&AcbVec> for AcbVec {
    type Output = AcbVec;

    fn add(mut self, rhs: &AcbVec) -> Self::Output {
        self += rhs;
        self
    }
}

impl Add<&AcbVec> for &AcbVec {
    type Output = AcbVec;

    fn add(self, rhs: &AcbVec) -> Self::Output {
        debug_assert!(self.len() == rhs.len(), "Lenghts do not match");
        let mut res = AcbVec::new(self.len());
        unsafe {
            acb_mat_add(&mut res.0, &self.0, &rhs.0, get_prec());
        }
        res
    }
}

// -
impl SubAssign<&AcbVec> for AcbVec {
    fn sub_assign(&mut self, rhs: &AcbVec) {
        debug_assert!(self.len() == rhs.len(), "Lenghts do not match");
        unsafe {
            acb_mat_sub(&mut self.0, &self.0, &rhs.0, get_prec());
        }
    }
}

impl Sub<AcbVec> for AcbVec {
    type Output = AcbVec;

    fn sub(mut self, rhs: AcbVec) -> Self::Output {
        self -= &rhs;
        self
    }
}

impl Sub<&AcbVec> for AcbVec {
    type Output = AcbVec;

    fn sub(mut self, rhs: &AcbVec) -> Self::Output {
        self -= rhs;
        self
    }
}

impl Sub<&AcbVec> for &AcbVec {
    type Output = AcbVec;

    fn sub(self, rhs: &AcbVec) -> Self::Output {
        debug_assert!(self.len() == rhs.len(), "Lenghts do not match");
        let mut res = AcbVec::new(self.len());
        unsafe {
            acb_mat_sub(&mut res.0, &self.0, &rhs.0, get_prec());
        }
        res
    }
}

impl Sub<AcbVec> for &AcbVec {
    type Output = AcbVec;

    fn sub(self, mut rhs: AcbVec) -> Self::Output {
        unsafe {
            acb_mat_sub(&mut rhs.0, &self.0, &rhs.0, get_prec());
        }
        rhs
    }
}

impl Neg for AcbVec {
    type Output = AcbVec;

    fn neg(mut self) -> Self::Output {
        unsafe {
            acb_mat_neg(&mut self.0, &self.0);
        }
        self
    }
}

// *
impl Mul<&Acb> for AcbVec {
    type Output = AcbVec;

    fn mul(mut self, rhs: &Acb) -> Self::Output {
        unsafe {
            acb_mat_scalar_mul_acb(&mut self.0, &self.0, &rhs.0, get_prec());
        }
        self
    }
}

impl Mul<Acb> for AcbVec {
    type Output = AcbVec;

    fn mul(self, rhs: Acb) -> Self::Output {
        self * &rhs
    }
}

impl Mul<&Acb> for &AcbVec {
    type Output = AcbVec;

    fn mul(self, rhs: &Acb) -> Self::Output {
        let mut res = AcbVec::new(self.len());
        unsafe {
            acb_mat_scalar_mul_acb(&mut res.0, &self.0, &rhs.0, get_prec());
        }
        res
    }
}

// /
impl Div<Acb> for AcbVec {
    type Output = AcbVec;

    fn div(mut self, rhs: Acb) -> Self::Output {
        (&mut self).into_iter().for_each(|c| *c /= &rhs);
        self
    }
}

// Vector
impl Vector for AcbVec {
    type T = Acb;

    fn from_value(nrows: usize, v: Self::T) -> Self {
        (0..nrows).map(|_| v.clone()).collect()
    }

    fn len(&self) -> usize {
        self.0.r.try_into().unwrap()
    }

    fn zeros(nrows: usize) -> Self
    where
        Self::T: num_traits::Zero,
    {
        AcbVec::new(nrows) // acb_mat_init initilizes a zero matrix
    }
}

// Interval
impl Interval for AcbVec {
    fn union(&self, other: &Self) -> Self {
        debug_assert!(self.len() == other.len(), "Vector sizes do not match");
        self.into_iter()
            .zip(other.into_iter())
            .map(|(self_i, other_i)| self_i.union(other_i))
            .collect()
    }

    fn diam(&self) -> f64 {
        self.into_iter()
            .map(|c| c.diam())
            .max_by(|x, y| x.total_cmp(y))
            .unwrap_or(0.0)
    }

    fn mag(&self) -> f64 {
        self.into_iter().map(|c| c.mag()).max_by(|x, y| x.total_cmp(y)).unwrap_or(0.0)
    }

    fn mag_lower(&self) -> f64 {
        self.into_iter()
            .map(|c| c.mag_lower())
            .max_by(|x, y| x.total_cmp(y))
            .unwrap_or(0.0)
    }

    fn mut_squash(&mut self) {
        self.into_iter().for_each(|c| c.mut_squash());
    }

    fn squash(&self) -> Self {
        self.into_iter().map(|x| x.squash()).collect()
    }

    fn square(rad: f64) -> Self {
        Self::from_value(1, Acb::square(rad))
    }
}
