use super::{AcbTm, AcbVec, Arb};
use crate::{
    box_vec::BoxVec,
    traits::{vector::Vector, TaylorModel},
};
use std::{
    fmt::Display,
    ops::{Add, AddAssign, Index, IndexMut, Mul},
};

#[derive(Clone, Debug)]
pub struct AcbVecTm<const N: usize>(pub BoxVec<AcbTm<N>>);

impl<const N: usize> Vector for AcbVecTm<N> {
    type T = AcbTm<N>;

    fn zeros(i: usize) -> Self
    where
        Self::T: num_traits::Zero,
    {
        Self(BoxVec::zeros(i))
    }

    fn from_value(i: usize, v: Self::T) -> Self {
        Self(BoxVec::from_value(i, v))
    }

    fn len(&self) -> usize {
        self.0.len()
    }

    fn get_unchecked_mut(&mut self, i: usize) -> &mut Self::T {
        self.0.get_unchecked_mut(i)
    }
}

impl<const N: usize> Display for AcbVecTm<N> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        self.0 .0.into_iter().map(|self_i| writeln!(f, "{}", self_i)).collect()
    }
}

impl<const N: usize> Index<usize> for AcbVecTm<N> {
    type Output = AcbTm<N>;

    fn index(&self, index: usize) -> &Self::Output {
        &self.0[index]
    }
}

impl<const N: usize> IndexMut<usize> for AcbVecTm<N> {
    fn index_mut(&mut self, index: usize) -> &mut Self::Output {
        &mut self.0[index]
    }
}

//TODO: maybe handle iterators in a better way to avoid res.0.0
impl<const N: usize> From<Vec<AcbTm<N>>> for AcbVecTm<N> {
    fn from(value: Vec<AcbTm<N>>) -> Self {
        let mut res = Self::zeros(value.len());
        res.0
             .0
            .iter_mut()
            .zip(value.into_iter())
            .for_each(|(res_i, value_i)| *res_i = value_i);
        res
    }
}

impl<const N: usize> AddAssign<&AcbVecTm<N>> for AcbVecTm<N> {
    fn add_assign(&mut self, rhs: &AcbVecTm<N>) {
        self.0 += &rhs.0;
    }
}

impl<const N: usize> TaylorModel for AcbVecTm<N> {
    type Coeffs = AcbVec;
    type Domain = Arb;

    fn order(&self) -> usize {
        N
    }

    fn from_vec(dom: &Self::Domain, coeffs: &Vec<Self::Coeffs>) -> Self {
        debug_assert!(!coeffs.is_empty());
        let n = coeffs[0].len();
        debug_assert!(coeffs.iter().all(|c| c.len() == n));
        let mut res: AcbVecTm<N> = AcbVecTm::zeros(n);
        for i in 0..res.len() {
            debug_assert!(coeffs.len() <= res[i].order(), "Coefficient vector is too large"); //TODO: improve this
            res[i].set_domain(dom);
            for (j, coeff_j) in coeffs.iter().enumerate() {
                res[i].set_coeff(coeff_j[i].clone(), j);
            }
        }
        res
    }

    fn domain(&self) -> Self::Domain {
        self.0.domain().clone()
    }

    fn set_domain(&mut self, dom: &Self::Domain) {
        self.0.set_domain(dom);
    }

    fn dominter(&mut self, dom: &Self::Domain) {
        self.0.dominter(dom);
    }

    fn enclosure(&self, dom: &Self::Domain) -> Self::Coeffs {
        self.0 .0.iter().map(|x| x.enclosure(dom)).collect()
    }

    fn set_coeff(&mut self, coeff: Self::Coeffs, i: usize) {
        debug_assert!(self.len() == coeff.len(), "Size error");
        for j in 0..self.len() {
            *self[j].get_coeff_mut(i) = coeff[j].clone() //TODO: remove clone
        }
    }

    fn get_coeff_mut(&mut self, _i: usize) -> &mut Self::Coeffs {
        unimplemented!()
    }

    fn affine(a: &Self::Coeffs, b: &Self::Coeffs, domain: &Self::Domain) -> Self {
        debug_assert!(a.len() == b.len(), "Lengths of vector coefficients mismatch");
        let mut a_auxi = BoxVec::zeros(a.len());
        let mut b_auxi = BoxVec::zeros(b.len());
        for i in 0..a.len() {
            a_auxi[i] = a[i].clone();
            b_auxi[i] = b[i].clone();
        }
        Self(BoxVec::affine(&a_auxi, &b_auxi, domain))
    }

    fn identity(domain: &Self::Domain) -> Self {
        Self(BoxVec::identity(domain))
    }
}

impl<const N: usize> From<&AcbVec> for AcbVecTm<N> {
    fn from(value: &AcbVec) -> Self {
        let mut res = BoxVec::zeros(value.len());
        for i in 0..value.len() {
            res[i] = value[i].clone().into()
        }
        Self(res)
    }
}

impl<const N: usize> Add for AcbVecTm<N> {
    type Output = Self;

    fn add(self, rhs: Self) -> Self::Output {
        Self(self.0 + rhs.0)
    }
}

impl<const N: usize> Mul<&AcbTm<N>> for AcbVecTm<N> {
    type Output = Self;

    fn mul(self, rhs: &AcbTm<N>) -> Self::Output {
        Self(self.0 * rhs)
    }
}

impl<const N: usize> Add<&AcbVecTm<N>> for AcbVecTm<N> {
    type Output = Self;

    fn add(self, rhs: &AcbVecTm<N>) -> Self::Output {
        Self(self.0 + &rhs.0)
    }
}
