use super::{Acb, Fmpq};
use crate::{
    precision_ctx::get_prec,
    reckless::Rif,
    traits::{
        interval::{IntervalComparisons, RealInterval},
        set::Set,
        Interval,
    },
};
use anyhow::bail;
use flint_bindings::*;
use num_traits::{Inv, One, Pow, Zero};
use std::{
    f64::consts::{LN_10, LN_2},
    ffi::{c_char, CStr},
    fmt::{Debug, Display},
    mem::MaybeUninit,
    ops::{Add, Mul, Sub},
    os::raw::c_void,
};

// Basics
pub struct Arb(pub arb_struct);

impl Default for Arb {
    fn default() -> Self {
        unsafe {
            let mut _new = MaybeUninit::uninit();
            arb_init(_new.as_mut_ptr());
            Self(_new.assume_init())
        }
    }
}

impl Clone for Arb {
    fn clone(&self) -> Self {
        let mut res = Arb::default();
        unsafe {
            arb_set(&mut res.0, &self.0);
        }
        res
    }
}

impl Drop for Arb {
    fn drop(&mut self) {
        unsafe {
            arb_clear(&mut self.0);
        }
    }
}

impl Debug for Arb {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        unsafe {
            let n = (f64::from(arb_bits(&self.0) as u32) * LN_2 / LN_10).ceil() as i64;
            let s_ptr = arb_get_str(&self.0, n, 1u64 + 2u64 * u64::from(self.diam() == 0.0)) as *mut c_char;
            if s_ptr.is_null() {
                return Err(std::fmt::Error);
            }
            let c_str = CStr::from_ptr(s_ptr);
            let result = match c_str.to_str() {
                Ok(s) => write!(f, "{}", s),
                Err(e) => {
                    write!(f, "{}", e)?;
                    Err(std::fmt::Error)
                }
            };
            flint_free(s_ptr as *mut c_void);
            result
        }
    }
}

impl Display for Arb {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{:?}", self)
    }
}

// Specific impls
impl Arb {
    pub fn try_get_included_reckless_interval(&self) -> Result<Rif, anyhow::Error> {
        unsafe {
            if arb_bits(&self.0) <= 53 && self.0.mid.exp <= 1023 && -1022 <= self.0.mid.exp {
                let res = arf_get_d(&self.0.mid, u32::MAX);
                let auxi = Arb::from(res);
                // Safety check
                if !auxi.subset(self) {
                    bail!("Conversion safety check failed");
                }
                Ok(res.into())
            } else {
                bail!("Exact conversion is impossible")
            }
        }
    }

    pub fn get_covering_reckless_interval(&self) -> Rif {
        unsafe {
            let mut auxi = MaybeUninit::uninit();
            arf_init(auxi.as_mut_ptr());
            let mut auxi = auxi.assume_init();
            arb_get_lbound_arf(&mut auxi, &self.0, 53);
            let lo = arf_get_d(&auxi, arf_rnd_t_ARF_RND_FLOOR);
            arb_get_ubound_arf(&mut auxi, &self.0, 53);
            let hi = arf_get_d(&auxi, arf_rnd_t_ARF_RND_CEIL);
            arf_clear(&mut auxi);
            Rif::new(lo, hi)
        }
    }

    pub fn _mag_round(&self, rnd: arf_rnd_t) -> f64 {
        let mut abs = Arb::default();
        let mut arf_rad = MaybeUninit::uninit();
        unsafe {
            arf_init(arf_rad.as_mut_ptr());
            let mut arf_rad = arf_rad.assume_init();
            arb_abs(&mut abs.0, &self.0);
            arf_set_mag(&mut arf_rad, &abs.0.rad);
            arf_add(&mut arf_rad, &abs.0.mid, &arf_rad, 53, rnd);
            let mag = arf_get_d(&arf_rad, rnd);
            arf_clear(&mut arf_rad);
            mag
        }
    }
}

// Special values and conversions
impl From<&arb_struct> for &Arb {
    fn from(value: &arb_struct) -> Self {
        unsafe { &*(value as *const arb_struct as *const Arb) }
    }
}

impl From<&mut arb_struct> for &mut Arb {
    fn from(value: &mut arb_struct) -> Self {
        unsafe { &mut *(value as *mut arb_struct as *mut Arb) }
    }
}

impl Zero for Arb {
    fn zero() -> Self {
        let mut res = Arb::default();
        unsafe {
            arb_zero(&mut res.0);
        }
        res
    }

    fn is_zero(&self) -> bool {
        unsafe { arb_is_zero(&self.0) != 0 }
    }
}

impl One for Arb {
    fn one() -> Self {
        let mut one = Arb::default();
        unsafe {
            arb_one(&mut one.0);
        }
        one
    }

    fn is_one(&self) -> bool
    where
        Self: PartialEq,
    {
        unsafe { arb_is_one(&self.0) != 0 }
    }

    fn set_one(&mut self) {
        unsafe { arb_one(&mut self.0) }
    }
}

impl From<f64> for Arb {
    fn from(value: f64) -> Self {
        let mut into = Arb::default();
        unsafe {
            arb_set_d(&mut into.0, value);
        }
        into
    }
}

impl From<Fmpq> for Arb {
    fn from(value: Fmpq) -> Self {
        let mut res = Arb::default();
        res.set(&value);
        res
    }
}

impl Set<&Fmpq> for Arb {
    fn set(&mut self, other: &Fmpq) {
        unsafe {
            arb_set_fmpq(&mut self.0, &other.0, get_prec());
        }
    }
}

impl TryFrom<&Acb> for Arb {
    type Error = anyhow::Error;

    fn try_from(value: &Acb) -> Result<Self, Self::Error> {
        if value.im().is_zero() {
            let mut res = Arb::default();
            unsafe {
                arb_set(&mut res.0, &value.0.real);
            }
            Ok(res)
        } else {
            bail!("Cannot convert complex interval to real interval with non-zero real part")
        }
    }
}

// Comparisons
impl PartialEq for Arb {
    fn eq(&self, other: &Self) -> bool {
        unsafe { arb_contains(&self.0, &other.0) != 0 && arb_contains(&other.0, &self.0) != 0 }
    }
}

impl Eq for Arb {}

impl IntervalComparisons for Arb {
    fn interval_ne(&self, other: &Self) -> bool {
        unsafe { arb_ne(&self.0, &other.0) != 0 }
    }

    fn interval_eq(&self, other: &Self) -> bool {
        unsafe { arb_eq(&self.0, &other.0) != 0 }
    }

    fn interval_ge(&self, other: &Self) -> bool {
        unsafe { arb_ge(&self.0, &other.0) != 0 }
    }

    fn interval_gt(&self, other: &Self) -> bool {
        unsafe { arb_gt(&self.0, &other.0) != 0 }
    }

    fn interval_le(&self, other: &Self) -> bool {
        unsafe { arb_le(&self.0, &other.0) != 0 }
    }

    fn interval_lt(&self, other: &Self) -> bool {
        unsafe { arb_lt(&self.0, &other.0) != 0 }
    }
}

// Arithmetic
impl Add<Arb> for Arb {
    type Output = Arb;

    fn add(self, rhs: Arb) -> Self::Output {
        self + &rhs
    }
}

impl Add<Arb> for &Arb {
    type Output = Arb;

    fn add(self, rhs: Arb) -> Self::Output {
        rhs + self
    }
}

impl Add<&Arb> for Arb {
    type Output = Arb;

    fn add(mut self, rhs: &Arb) -> Self::Output {
        unsafe {
            arb_add(&mut self.0, &self.0, &rhs.0, get_prec());
        }
        self
    }
}

impl Mul<Arb> for Arb {
    type Output = Arb;

    fn mul(mut self, rhs: Arb) -> Self::Output {
        unsafe {
            arb_mul(&mut self.0, &self.0, &rhs.0, get_prec());
        }
        self
    }
}

impl Inv for Arb {
    type Output = Arb;

    fn inv(mut self) -> Self::Output {
        unsafe {
            arb_inv(&mut self.0, &self.0, get_prec());
        }
        self
    }
}

impl Sub<&Arb> for Arb {
    type Output = Arb;

    fn sub(mut self, rhs: &Arb) -> Self::Output {
        unsafe {
            arb_sub(&mut self.0, &self.0, &rhs.0, get_prec());
        }
        self
    }
}

impl Sub<&Arb> for &Arb {
    type Output = Arb;

    fn sub(self, rhs: &Arb) -> Self::Output {
        let mut res = Arb::default();
        unsafe {
            arb_sub(&mut res.0, &self.0, &rhs.0, get_prec());
        }
        res
    }
}

impl Sub<Arb> for Arb {
    type Output = Arb;

    fn sub(self, rhs: Arb) -> Self::Output {
        self - &rhs
    }
}

impl Pow<usize> for &Arb {
    type Output = Arb;

    fn pow(self, rhs: usize) -> Self::Output {
        let mut res = Arb::default();
        unsafe {
            arb_pow_ui(&mut res.0, &self.0, rhs as u64, get_prec());
        }
        res
    }
}

// Interval
impl RealInterval for Arb {
    fn entire() -> Self {
        let mut entire = Arb::default();
        unsafe {
            arb_zero_pm_inf(&mut entire.0);
        }
        entire
    }

    fn from_inf_sup(inf: Arb, sup: Arb) -> Self {
        let mut res = Arb::default();
        unsafe {
            let mut _inf_arf = MaybeUninit::uninit();
            let mut _sup_arf = MaybeUninit::uninit();
            arf_init(_inf_arf.as_mut_ptr());
            arf_init(_sup_arf.as_mut_ptr());
            let mut inf_arf = _inf_arf.assume_init();
            let mut sup_arf = _sup_arf.assume_init();

            arb_get_lbound_arf(&mut inf_arf, &inf.0, get_prec()); // inf_arf is a lower bound for inf
            arb_get_ubound_arf(&mut sup_arf, &sup.0, get_prec()); // sup_arf is a lower bound for sup

            arb_set_interval_arf(&mut res.0, &inf_arf, &sup_arf, get_prec());
            arf_clear(&mut inf_arf);
            arf_clear(&mut sup_arf);
        }
        res
    }

    fn inf(&self) -> Arb {
        let mut res = Arb::default();
        unsafe {
            let mut _inf = MaybeUninit::uninit();
            arf_init(_inf.as_mut_ptr());
            let mut inf = _inf.assume_init();
            arb_get_lbound_arf(&mut inf, &self.0, get_prec());
            arb_set_arf(&mut res.0, &inf);
            arf_clear(&mut inf);
        }
        res
    }

    fn intersection(&self, other: &Self) -> Self {
        let mut res = Arb::default();
        unsafe {
            arb_intersection(&mut res.0, &self.0, &other.0, get_prec());
        }
        res
    }

    fn sup(&self) -> Arb {
        let mut res = Arb::default();
        unsafe {
            let mut _sup = MaybeUninit::uninit();
            arf_init(_sup.as_mut_ptr());
            let mut sup = _sup.assume_init();
            arb_get_ubound_arf(&mut sup, &self.0, get_prec());
            arb_set_arf(&mut res.0, &sup);
            arf_clear(&mut sup);
        }
        res
    }

    fn subset(&self, other: &Self) -> bool {
        unsafe { arb_contains(&other.0, &self.0) != 0 }
    }
}

impl Interval for Arb {
    fn union(&self, other: &Self) -> Self {
        let mut res = Arb::default();
        unsafe {
            arb_union(&mut res.0, &self.0, &other.0, get_prec());
        }
        res
    }

    fn diam(&self) -> f64 {
        unsafe { mag_get_d(&self.0.rad) * 2.0 } // hopefully this operation is exact
                                                // unsafe { with_rounding_up(|| mag_get_d(&self.0.rad) * 2.0) } // safer version
    }

    fn mag(&self) -> f64 {
        self._mag_round(arf_rnd_t_ARF_RND_CEIL)
    }

    fn mag_lower(&self) -> f64 {
        self._mag_round(arf_rnd_t_ARF_RND_FLOOR)
    }

    fn mut_squash(&mut self) {
        unsafe {
            arb_get_mid_arb(&mut self.0, &self.0);
        }
    }

    fn square(rad: f64) -> Self {
        let mut res = Arb::default();
        unsafe {
            mag_set_d(&mut res.0.rad, rad);
            arf_zero(&mut res.0.mid); //TODO: test to ensure this is useless
        }
        res
    }

    fn squash(&self) -> Self {
        let mut mid = Arb::default();
        unsafe {
            arb_get_mid_arb(&mut mid.0, &self.0);
        }
        mid
    }
}
