use anyhow::bail;
use flint_bindings::*;
use std::str::FromStr;
use std::{
    ffi::CString,
    mem::MaybeUninit,
    ops::{Add, AddAssign, Div, Mul, Neg},
};

pub struct Fmpq(pub fmpq);

impl Default for Fmpq {
    fn default() -> Self {
        unsafe {
            let mut new = MaybeUninit::uninit();
            fmpq_init(new.as_mut_ptr());
            Self(new.assume_init())
        }
    }
}

impl Fmpq {
    pub fn from_machine_int(numer: i64, denom: u64) -> Self {
        let mut res = Fmpq::default();
        unsafe {
            fmpq_set_si(&mut res.0, numer, denom);
        }
        res
    }
}

impl Drop for Fmpq {
    fn drop(&mut self) {
        unsafe {
            fmpq_clear(&mut self.0);
        }
    }
}

impl From<&fmpq> for &Fmpq {
    fn from(src: &fmpq) -> Self {
        unsafe { &*(src as *const fmpq as *const Fmpq) }
    }
}

impl From<&mut fmpq> for &mut Fmpq {
    fn from(src: &mut fmpq) -> Self {
        unsafe { &mut *(src as *mut fmpq as *mut Fmpq) }
    }
}

impl FromStr for Fmpq {
    type Err = anyhow::Error;

    fn from_str(s: &str) -> Result<Fmpq, Self::Err> {
        let mut r = Fmpq::default();
        let c_string = CString::new(s).unwrap();
        let res = unsafe { fmpq_set_str(&mut r.0, c_string.as_ptr(), 10) };
        if res == 0 {
            Ok(r)
        } else {
            bail!("Cannot read rational")
        }
    }
}

impl From<i64> for Fmpq {
    fn from(i: i64) -> Fmpq {
        let mut r = Fmpq::default();
        unsafe {
            fmpq_set_si(&mut r.0, i, 1);
        }
        r
    }
}

impl Clone for Fmpq {
    fn clone(&self) -> Self {
        let mut res = Fmpq::default();
        unsafe {
            fmpq_set(&mut res.0, &self.0);
        }
        res
    }
}

impl std::fmt::Display for Fmpq {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        unsafe {
            let s_ptr = fmpq_get_str(std::ptr::null_mut(), 10, &self.0) as *mut std::ffi::c_char;
            if s_ptr.is_null() {
                return Err(std::fmt::Error);
            }
            let c_str = std::ffi::CStr::from_ptr(s_ptr);
            let result = match c_str.to_str() {
                Ok(s) => write!(f, "{}", s),
                Err(e) => {
                    write!(f, "{}", e)?;
                    Err(std::fmt::Error)
                }
            };
            flint_free(s_ptr as *mut std::ffi::c_void);
            result
        }
    }
}

impl std::fmt::Debug for Fmpq {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        std::fmt::Display::fmt(self, f)
    }
}

impl PartialEq for Fmpq {
    fn eq(&self, other: &Self) -> bool {
        unsafe { fmpq_equal(&self.0, &other.0) != 0 }
    }
}

impl Eq for Fmpq {}

impl std::hash::Hash for Fmpq {
    fn hash<H>(&self, state: &mut H)
    where
        H: std::hash::Hasher,
    {
        unsafe {
            state.write_u64(calcium_fmpz_hash(&self.0.den));
            state.write_u64(calcium_fmpz_hash(&self.0.num));
        }
    }
}

impl Add<Fmpq> for Fmpq {
    type Output = Self;

    fn add(mut self, rhs: Fmpq) -> Self::Output {
        unsafe {
            fmpq_add(&mut self.0, &self.0, &rhs.0);
        }
        self
    }
}

impl std::ops::Sub<Fmpq> for Fmpq {
    type Output = Self;

    fn sub(mut self, rhs: Fmpq) -> Self::Output {
        unsafe {
            fmpq_sub(&mut self.0, &self.0, &rhs.0);
        }
        self
    }
}

impl Mul<Fmpq> for Fmpq {
    type Output = Self;

    fn mul(mut self, rhs: Fmpq) -> Self::Output {
        unsafe {
            fmpq_mul(&mut self.0, &self.0, &rhs.0);
        }
        self
    }
}

impl Fmpq {
    pub fn pow(&self, exp: i64) -> Self {
        let mut r = Fmpq::default();
        unsafe {
            fmpq_pow_si(&mut r.0, &self.0, exp);
        }
        r
    }
}

impl Div<Fmpq> for Fmpq {
    type Output = Self;

    fn div(mut self, rhs: Fmpq) -> Self::Output {
        unsafe {
            fmpq_div(&mut self.0, &self.0, &rhs.0);
        }
        self
    }
}

impl num_traits::One for Fmpq {
    fn one() -> Self {
        let mut one = Fmpq::default();
        one.set_one();
        one
    }

    fn set_one(&mut self) {
        unsafe {
            fmpq_one(&mut self.0);
        }
    }

    fn is_one(&self) -> bool
    where
        Self: PartialEq,
    {
        unsafe { fmpq_is_one(&self.0) != 0 }
    }
}

impl num_traits::Zero for Fmpq {
    fn zero() -> Self {
        let mut zero = Fmpq::default();
        zero.set_zero();
        zero
    }

    fn set_zero(&mut self) {
        unsafe {
            fmpq_zero(&mut self.0);
        }
    }

    fn is_zero(&self) -> bool
    where
        Self: PartialEq,
    {
        unsafe { fmpq_is_zero(&self.0) != 0 }
    }
}

impl Neg for Fmpq {
    type Output = Self;

    fn neg(mut self) -> Self::Output {
        unsafe {
            fmpq_neg(&mut self.0, &self.0);
        }
        self
    }
}

impl AddAssign<&Fmpq> for Fmpq {
    fn add_assign(&mut self, rhs: &Self) {
        unsafe {
            fmpq_add(&mut self.0, &self.0, &rhs.0);
        }
    }
}

// Only useful to implement num_traits::Num
impl std::ops::Rem for Fmpq {
    type Output = Self;
    fn rem(self, _rhs: Self) -> Self {
        Fmpq::default()
    }
}

impl num_traits::Num for Fmpq {
    type FromStrRadixErr = ();
    fn from_str_radix(s: &str, radix: u32) -> Result<Self, ()> {
        let mut r = Fmpq::default();
        let c_string = CString::new(s).unwrap();
        let res = unsafe { fmpq_set_str(&mut r.0, c_string.as_ptr(), radix.try_into().unwrap()) };
        if res == 0 {
            Ok(r)
        } else {
            Err(())
        }
    }
}

#[cfg(test)]
mod test {
    use super::*;
    #[test]
    fn read_display() {
        let in1 = "1234/5678";
        let r = Fmpq::from_str(in1).unwrap();
        assert_eq!(format!("{}", r), in1);

        let in2 =
            "12342130129831209381203981203912045678982311/1232139871239874535348567345673458712836123876123237542386128376135678";
        let r = Fmpq::from_str(in2).unwrap();
        assert_eq!(format!("{}", r), in2);
    }

    #[test]
    fn hash_map() {
        let in1 = "1234/5678";
        let r1 = Fmpq::from_str(in1).unwrap();

        let in2 =
            "12342130129831209381203981203912045678982311/1232139871239874535348567345673458712836123876123237542386128376135678";
        let r2 = Fmpq::from_str(in2).unwrap();

        let mut h = std::collections::HashMap::new();
        h.insert(r1.clone(), 1usize);
        h.insert(r2, 2);

        assert_eq!(h.get(&r1).cloned(), Some(1));
    }

    #[test]
    fn circuit_over_complex_rational() {
        use std::sync::Arc;
        let ctx = crate::circuit::Context::<num_complex::Complex<Fmpq>>::default();
        let x = ctx.var("x");
        let y = ctx.var("y");

        let a = ctx.scalar(num_complex::Complex::new(
            Fmpq::from_str("2").unwrap().into(),
            Fmpq::from_str("4/7").unwrap().into(),
        ));

        let z = x.clone() + a.clone() * &y + x.clone() * &y;
        let zbis = x.clone() + a * &y + x.clone() * &y;
        assert!(Arc::ptr_eq(&z.node, &zbis.node));

        let ybis = ctx.var("y");
        assert!(Arc::ptr_eq(&y.node, &ybis.node));
    }
}
