use std::{fmt::Display, ops::{Add, Index, IndexMut}};

use nalgebra::Complex;
use num_traits::Zero;

use crate::{
    box_mat::BoxMat,
    box_vec::BoxVec,
    flint::{Acb, AcbMat, AcbVec, Arb, Fmpq},
    reckless::{Cif, Rif},
    traits::{
        interval::{IntervalComparisons, RealInterval},
        matrix::Matrix,
        vector::Vector,
        Interval,
    },
};

#[derive(Clone, Debug)]
pub enum FoA<F, A> {
    Fast(F),
    Accurate(A),
}

impl<F, A> Display for FoA<F, A> where F: Display, A: Display {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            FoA::Fast(fast) => write!(f, "{}", fast),
            FoA::Accurate(accurate) => write!(f, "{}", accurate),
        }
    }
}

pub trait FoATrait {
    type F;
    type A;

    fn get_fast(self) -> anyhow::Result<Self::F>;
    fn ref_get_fast(&self) -> anyhow::Result<Self::F>;
    // fn get_fast_sloppy(self) -> Self::F;
    // fn gety_fast_sloppy(self) -> Self::F;
    fn get_accurate(self) -> Self::A;
    fn ref_get_accurate(&self) -> Self::A;
}

pub trait FaATrait {
    type F;
    type A;

    fn get_fast(&self) -> &Self::F;
    fn get_accurate(&self) -> &Self::A;
}

impl FoATrait for FoA<BoxVec<Cif>, AcbVec> {
    type F = BoxVec<Cif>;
    type A = AcbVec;

    fn get_fast(self) -> anyhow::Result<Self::F> {
        todo!()
    }

    fn ref_get_fast(&self) -> anyhow::Result<Self::F> {
        match self {
            FoA::Fast(fast) => Ok(fast.clone()),
            FoA::Accurate(accurate) => {
                let auxi = accurate
                    .into_iter()
                    .map(|c| c.try_get_included_reckless_interval())
                    .collect::<anyhow::Result<Vec<Cif>>>()?;
                Ok(auxi.into())
            }
        }
    }

    fn get_accurate(self) -> Self::A {
        match self {
            FoA::Accurate(accurate) => accurate,
            FoA::Fast(fast) => fast.0.iter().map(|c| c.get_included_acb()).collect::<Vec<Acb>>().into(),
        }
    }

    fn ref_get_accurate(&self) -> Self::A {
        match self {
            FoA::Accurate(accurate) => accurate.clone(),
            FoA::Fast(fast) => fast.0.iter().map(|c| c.get_included_acb()).collect::<Vec<Acb>>().into(),
        }
    }
}

impl FoATrait for FoA<BoxMat<Cif>, AcbMat> {
    type F = BoxMat<Cif>;
    type A = AcbMat;

    fn get_fast(self) -> anyhow::Result<Self::F> {
        todo!()
    }

    fn ref_get_fast(&self) -> anyhow::Result<Self::F> {
        match self {
            FoA::Fast(fast) => Ok(fast.clone()),
            FoA::Accurate(accurate) => {
                let mut mat = BoxMat::zeros(accurate.ncols(), accurate.nrows());
                for i in 0..mat.nrows() {
                    for j in 0..mat.ncols() {
                        mat[(i, j)] = accurate[(i, j)].try_get_included_reckless_interval()?;
                    }
                }
                Ok(mat)
            }
        }
    }

    fn get_accurate(self) -> Self::A {
        match self {
            FoA::Accurate(accurate) => accurate,
            FoA::Fast(fast) => {
                let mut mat = AcbMat::zeros(fast.ncols(), fast.nrows());
                for i in 0..mat.nrows() {
                    for j in 0..mat.ncols() {
                        mat[(i, j)] = fast[(i, j)].get_included_acb();
                    }
                }
                mat
            }
        }
    }

    fn ref_get_accurate(&self) -> Self::A {
        todo!()
    }
}

impl FoATrait for FoA<Cif, Acb> {
    type F = Cif;
    type A = Acb;

    fn get_accurate(self) -> Self::A {
        match self {
            FoA::Accurate(accurate) => accurate,
            FoA::Fast(fast) => Cif::get_included_acb(&fast),
        }
    }

    fn get_fast(self) -> anyhow::Result<Self::F> {
        todo!()
    }

    fn ref_get_accurate(&self) -> Self::A {
        todo!()
    }

    fn ref_get_fast(&self) -> anyhow::Result<Self::F> {
        todo!()
    }
}

impl FoATrait for FoA<Rif, Arb> {
    type A = Arb;
    type F = Rif;

    fn get_accurate(self) -> Self::A {
        match self {
            FoA::Accurate(accurate) => accurate,
            FoA::Fast(fast) => fast.get_included_arb(),
        }
    }

    fn get_fast(self) -> anyhow::Result<Self::F> {
        todo!()
    }

    fn ref_get_accurate(&self) -> Self::A {
        match self {
            FoA::Accurate(accurate) => accurate.clone(),
            FoA::Fast(fast) => fast.get_included_arb(),
        }
    }

    fn ref_get_fast(&self) -> anyhow::Result<Self::F> {
        match self {
            FoA::Fast(fast) => Ok(fast.clone()),
            FoA::Accurate(accurate) => accurate.try_get_included_reckless_interval(),
        }
    }
}

impl Index<usize> for FoA<BoxVec<Cif>, AcbVec> {
    type Output = FoA<Cif, Acb>;

    fn index(&self, index: usize) -> &Self::Output {
        unimplemented!()
    }
}

impl IndexMut<usize> for FoA<BoxVec<Cif>, AcbVec> {
    fn index_mut(&mut self, index: usize) -> &mut Self::Output {
        unimplemented!()
    }
}

impl From<Vec<FoA<Cif, Acb>>> for FoA<BoxVec<Cif>, AcbVec> {
    fn from(value: Vec<FoA<Cif, Acb>>) -> Self {
        let v: Vec<Acb> = value.into_iter().map(|c| c.get_accurate()).collect();
        FoA::Accurate(AcbVec::from(v))
    }
}

impl Vector for FoA<BoxVec<Cif>, AcbVec> {
    type T = FoA<Cif, Acb>;

    fn len(&self) -> usize {
        match self {
            FoA::Fast(f) => f.len(),
            FoA::Accurate(a) => a.len(),
        }
    }

    fn from_value(i: usize, v: Self::T) -> Self {
        unimplemented!()
    }

    fn get_unchecked_mut(&mut self, i: usize) -> &mut Self::T {
        unimplemented!()
    }

    fn zeros(i: usize) -> Self {
        unimplemented!()
    }
}

impl Interval for FoA<BoxVec<Cif>, AcbVec> {
    fn diam(&self) -> f64 {
        unimplemented!()
    }

    fn mag(&self) -> f64 {
        match self {
            FoA::Fast(fast) => fast.mag(),
            FoA::Accurate(accurate) => accurate.mag(),
        }
    }

    fn mag_lower(&self) -> f64 {
        unimplemented!()
    }

    fn mut_squash(&mut self) {
        unimplemented!()
    }

    fn square(rad: f64) -> Self {
        unimplemented!()
    }

    fn squash(&self) -> Self {
        unimplemented!()
    }

    fn union(&self, other: &Self) -> Self {
        unimplemented!()
    }
}

impl Add for FoA<Cif, Acb> {
    type Output = FoA<Cif, Acb>;

    fn add(self, rhs: Self) -> Self::Output {
        unimplemented!()
    }
}

impl Zero for FoA<Cif, Acb> {
    fn zero() -> Self {
        FoA::Fast(Cif::zero())
    }

    fn is_zero(&self) -> bool {
        match self {
            FoA::Fast(fast) => fast.is_zero(),
            FoA::Accurate(accurate) => accurate.is_zero(),
        }
    }
}

impl TryInto<FoA<Rif, Arb>> for &FoA<Cif, Acb> {
    type Error = anyhow::Error;

    fn try_into(self) -> Result<FoA<Rif, Arb>, Self::Error> {
        todo!()
    }
}

impl From<f64> for FoA<Rif, Arb> {
    fn from(value: f64) -> Self {
        FoA::Accurate(Arb::from(value))
    }
}

impl Interval for FoA<Rif, Arb> {
    fn diam(&self) -> f64 {
        todo!()
    }

    fn mag(&self) -> f64 {
        todo!()
    }

    fn mag_lower(&self) -> f64 {
        todo!()
    }

    fn mut_squash(&mut self) {
        todo!()
    }

    fn square(rad: f64) -> Self {
        todo!()
    }

    fn squash(&self) -> Self {
        todo!()
    }

    fn union(&self, other: &Self) -> Self {
        todo!()
    }
}

impl RealInterval for FoA<Rif, Arb> {
    fn entire() -> Self {
        todo!()
    }

    fn from_inf_sup(inf: Self, sup: Self) -> Self {
        todo!()
    }

    fn inf(&self) -> Self {
        todo!()
    }

    fn intersection(&self, other: &Self) -> Self {
        todo!()
    }

    fn subset(&self, other: &Self) -> bool {
        todo!()
    }

    fn sup(&self) -> Self {
        todo!()
    }
}

impl IntervalComparisons for FoA<Rif, Arb> {
    fn interval_eq(&self, other: &Self) -> bool {
        todo!()
    }

    fn interval_ge(&self, other: &Self) -> bool {
        todo!()
    }

    fn interval_gt(&self, other: &Self) -> bool {
        todo!()
    }

    fn interval_le(&self, other: &Self) -> bool {
        todo!()
    }

    fn interval_lt(&self, other: &Self) -> bool {
        match (self, other) {
            (FoA::Fast(self_fast), FoA::Fast(other_fast)) => self_fast.interval_lt(other_fast),
            (FoA::Accurate(self_accurate), FoA::Accurate(other_accurate)) => self_accurate.interval_lt(other_accurate),
            (FoA::Fast(_), FoA::Accurate(other_accurate)) => self.ref_get_accurate().interval_lt(other_accurate),
            (FoA::Accurate(self_accurate), FoA::Fast(_)) => self_accurate.interval_lt(&other.ref_get_accurate()),
        }
    }

    fn interval_ne(&self, other: &Self) -> bool {
        todo!()
    }
}

// for<'r> TrackAlgo::C: From<&'r Complex<Fmpq>>,
impl From<&Complex<Fmpq>> for FoA<Cif, Acb> {
    fn from(value: &Complex<Fmpq>) -> Self {
        FoA::Accurate(Acb::from(value))
    }
}

impl Add for FoA<Rif, Arb> {
    type Output = FoA<Rif, Arb>;

    fn add(self, rhs: Self) -> Self::Output {
        todo!()
    }
}

impl Zero for FoA<Rif, Arb> {
    fn zero() -> Self {
        FoA::Fast(Rif::zero())
    }

    fn is_zero(&self) -> bool {
        match self {
            FoA::Fast(fast) => fast.is_zero(),
            FoA::Accurate(accurate) => accurate.is_zero(),
        }
    }

    fn set_zero(&mut self) {
        todo!()
    }
}
