use crate::{
    circuit::{Circuit, Context},
    flint::Fmpq,
};
use anyhow::bail;
use num_complex::Complex;
use num_traits::Zero;
use std::str::FromStr;

#[derive(Debug, Clone, PartialEq, Eq)]
pub enum Node {
    Add(Tree, Tree),
    Mul(Tree, Tree),
    Neg(Tree),
    Power(Tree, i64),
    Float(FloatFormat),
    Var(String),
    Int(String),
    Inverse(Tree),
    ImaginaryUnit,
}

pub type Tree = Box<Node>;

#[derive(PartialEq, Eq, Debug, Clone)]
pub struct FloatFormat {
    pub beforedot: String,
    pub afterdot: Option<String>,
    pub exponent: Option<String>,
}

impl TryFrom<&str> for FloatFormat {
    type Error = ();
    fn try_from(s: &str) -> Result<FloatFormat, ()> {
        let mut expsplit = s.split('e');

        let beforeexp = expsplit.next().ok_or(())?;
        let exponent = expsplit.next();

        let mut dotsplit = beforeexp.split('.');
        let beforedot = dotsplit.next().ok_or(())?;
        let afterdot = dotsplit.next();

        Ok(FloatFormat {
            beforedot: beforedot.into(),
            afterdot: afterdot.map(|s| s.into()),
            exponent: exponent.map(|s| s.into()),
        })
    }
}

impl Node {
    pub fn into_circuit(
        &self,
        ctx: &Context<Complex<Fmpq>>,
    ) -> anyhow::Result<Circuit<Complex<Fmpq>>> {
        use Node::*;
        match self {
            Add(s, t) => Ok(s.into_circuit(ctx)? + t.into_circuit(ctx)?),
            Mul(s, t) => Ok(s.into_circuit(ctx)? * t.into_circuit(ctx)?),
            Neg(t) => Ok(-t.into_circuit(ctx)?),
            Power(t, n) => Ok(t.into_circuit(ctx)?.pow(*n)),

            Float(f) => {
                let mut a = Fmpq::from_str(&f.beforedot)?;
                let ten: Fmpq = 10i64.into();
                if let Some(frac) = &f.afterdot {
                    let b = Fmpq::from_str(frac)?;
                    let mult = ten.pow(-(frac.len() as i64));
                    a += &(b * mult);
                }

                if let Some(exp) = &f.exponent {
                    let exp = i64::from_str(exp)?;
                    a = a * ten.pow(exp);
                }

                Ok(ctx.scalar(Complex::new(a, Fmpq::zero())))
            }

            Var(name) => Ok(ctx.var(name)),

            Int(istr) => {
                let a = Fmpq::from_str(istr)?;
                Ok(ctx.scalar(Complex::new(a, Fmpq::zero())))
            }

            Inverse(t) => {
                let scalar = t.into_circuit(ctx)?;
                match scalar.get_scalar() {
                    Some(s) => Ok(ctx.scalar(s.inv())),
                    None => bail!("Cannot inverse a nonscalar polynomial"),
                }
            }

            ImaginaryUnit => Ok(ctx.scalar(Complex::i())),
        }
    }
}

#[test]
fn test_float_format() {
    assert_eq!(
        "1".try_into(),
        Ok(FloatFormat {
            beforedot: "1".into(),
            afterdot: None,
            exponent: None
        })
    );
    assert_eq!(
        "1.000".try_into(),
        Ok(FloatFormat {
            beforedot: "1".into(),
            afterdot: Some("000".into()),
            exponent: None
        })
    );
    assert_eq!(
        "1.000".try_into(),
        Ok(FloatFormat {
            beforedot: "1".into(),
            afterdot: Some("000".into()),
            exponent: None
        })
    );
    assert_eq!(
        "1231.000".try_into(),
        Ok(FloatFormat {
            beforedot: "1231".into(),
            afterdot: Some("000".into()),
            exponent: None
        })
    );
    assert_eq!(
        "1231e-10".try_into(),
        Ok(FloatFormat {
            beforedot: "1231".into(),
            afterdot: None,
            exponent: Some("-10".into())
        })
    );
}

#[test]
fn test_into_circuit() {
    let ast = super::POLY
        .parse("x + (200*I + y)^2*3 - 4.1e-100*x/6")
        .unwrap();
    let mut ctx = Context::<Complex<Fmpq>>::default();

    let circuit = ast.into_circuit(&mut ctx).unwrap();
    let x = ctx.var("x");
    let y = ctx.var("y");

    let mut pt = std::collections::HashMap::default();
    pt.insert(x.clone(), Fmpq::from_machine_int(3, 1).into());
    pt.insert(y.clone(), Fmpq::from_machine_int(6, 1).into());

    let re = "-23977800000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000041/200000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000";
    assert_eq!(
        circuit.eval(&pt),
        Some(Complex::new(
            Fmpq::from_str(re).unwrap(),
            Fmpq::from_str("7200").unwrap()
        ))
    );
}
