// Copyright 2024 Pierre Lairez
// SPDX-License-Identifier: GPL-3.0-or-later

use crate::{
    box_mat::BoxMat,
    box_vec::BoxVec,
    flint::{Acb, AcbMat, AcbVec, Arb, Fmpq},
    prog::Eval,
    reckless::{Cif, Rif},
    traits::{matrix::Matrix, num::InvMat, vector::Vector, Interval},
};
use nalgebra::Complex;
use num_traits::{One, Zero};
use std::{
    cmp::Ordering,
    fmt::Debug,
    ops::{Div, SubAssign},
};

pub type ExactSystem = ParametricSystem<Complex<Fmpq>>;

#[derive(Clone)]
pub struct ParametricSystem<Prog> {
    pub nparams: usize,
    pub dim: usize,
    pub chart_switch_threshold: f64,
    pub f: Vec<Prog>,
    pub fdot: Vec<Prog>,
    pub df: Vec<Prog>,
}

#[derive(Debug, Clone)]
pub struct ProjectivePoint<VecT> {
    pub coord: VecT,
    pub chart: usize,
}

impl<VecT> SubAssign<&VecT> for ProjectivePoint<VecT>
where
    VecT: Matrix,
    <VecT as Matrix>::Coeffs: for<'a> SubAssign<&'a <VecT as Matrix>::Coeffs>,
{
    fn sub_assign(&mut self, delta: &VecT) {
        let mut offset = 0;
        for i in 0..self.coord.nrows() {
            if i == self.chart {
                offset = 1;
            } else {
                self.coord[(i, 0)] -= &delta[(i - offset, 0)];
            }
        }
    }
}

impl<VecT: Vector> ProjectivePoint<VecT> {
    fn projective_coordinate_mut(&mut self, i: usize) -> Option<&mut <VecT as Vector>::T> {
        match i.cmp(&self.chart) {
            Ordering::Equal => None,
            Ordering::Less => Some(&mut self.coord[i]),
            _ => Some(&mut self.coord[i - 1]),
        }
    }
}

impl<VecT> ProjectivePoint<VecT>
where
    VecT: Vector,
    VecT::T: One + Clone,
{
    pub fn projective_coordinate(&self, i: usize) -> <VecT as Vector>::T {
        match i.cmp(&self.chart) {
            Ordering::Equal => <VecT as Vector>::T::one(),
            Ordering::Less => self.coord[i].clone(),
            _ => self.coord[i - 1].clone(),
        }
    }
}

impl<VecT> ProjectivePoint<VecT>
where
    VecT: Vector,
    VecT::T: Zero + One + Clone + Interval,
{
    pub fn canonical_chart(&self) -> usize {
        let mut chart = 0;
        let mut maxmag = 1.0;
        for i in 0..=self.coord.len() {
            let c = self.projective_coordinate(i);
            if c.mag() > maxmag {
                chart = i;
                maxmag = c.mag();
            }
        }

        chart
    }
}

impl<VecT> ProjectivePoint<VecT>
where
    VecT: Vector + Debug,
    <VecT as Vector>::T: Zero + One + Interval,
    for<'a> <VecT as Vector>::T: Div<&'a <VecT as Vector>::T, Output = <VecT as Vector>::T>,
{
    pub fn to_canonical_chart(&self) -> Self {
        let chart = self.canonical_chart();
        self.to_chart(chart)
    }

    pub fn to_chart(&self, chart: usize) -> Self {
        let ccoeff = self.projective_coordinate(chart);
        let mag = ccoeff.mag();
        let mut pt = ProjectivePoint {
            coord: VecT::zeros(self.coord.len()),
            chart,
        };

        if mag.is_one() {
            pt.coord = self.coord.clone();
            return pt;
        }

        for i in 0..=pt.coord.len() {
            if let Some(c) = pt.projective_coordinate_mut(i) {
                *c = self.projective_coordinate(i) / &ccoeff;
            }
        }

        pt
    }
}

pub trait InvDf<VecS, MatS> {
    type S;

    fn inv_df(&self, params: &Self::S, pt: &ProjectivePoint<VecS>) -> anyhow::Result<MatS>;
}

impl<Prog, VecS, MatS> InvDf<VecS, MatS> for ParametricSystem<Prog>
where
    VecS: Vector,
    VecS::T: Zero + Clone,
    for<'r1, 'r2> Prog: Eval<(&'r1 VecS::T, &'r2 VecS), Vec<VecS::T>>,
    MatS: Matrix<Coeffs = VecS::T> + InvMat + Interval,
{
    type S = VecS::T;

    fn inv_df(&self, params: &Self::S, pt: &ProjectivePoint<VecS>) -> anyhow::Result<MatS> {
        let df: MatS = self.df::<VecS, MatS>(params, pt).squash();
        MatS::inv_mat(&df)
    }
}

pub trait FwBuf<VecS> {
    type S;

    fn f(&self, params: &Self::S, pt: &ProjectivePoint<VecS>) -> VecS;
}

impl<Prog, VecS> FwBuf<VecS> for ParametricSystem<Prog>
where
    VecS::T: Zero + Clone,
    VecS: Vector,
    for<'r1, 'r2> Prog: Eval<(&'r1 VecS::T, &'r2 VecS), Vec<VecS::T>>,
{
    type S = VecS::T;

    fn f(&self, params: &Self::S, pt: &ProjectivePoint<VecS>) -> VecS {
        self.f[pt.chart].eval((params, &pt.coord)).into()
    }
}

impl<Prog> ParametricSystem<Prog> {
    pub fn df<VecS, MatS>(&self, params: &VecS::T, pt: &ProjectivePoint<VecS>) -> MatS
    where
        MatS::Coeffs: Zero,
        MatS: Matrix,
        VecS: Vector<T = MatS::Coeffs>,
        for<'r1, 'r2> Prog: Eval<(&'r1 VecS::T, &'r2 VecS), Vec<VecS::T>>,
    {
        MatS::from_vec(self.dim, self.dim, self.df[pt.chart].eval((params, &pt.coord)))
    }
}

pub type CifMb = MooreBox<Rif, BoxVec<Cif>, BoxMat<Cif>>;

#[derive(Debug, Clone)]
pub struct MooreBox<R, VecC, MatC> {
    pub params: R,
    pub pt: ProjectivePoint<VecC>,
    pub mat: MatC,
    pub rad: f64,
}

impl<R, VecC, MatC> MooreBox<R, VecC, MatC> {
    pub fn mut_squash(&mut self)
    where
        R: Interval,
        VecC: Interval,
        MatC: Interval,
    {
        self.params.mut_squash();
        self.pt.coord.mut_squash();
        self.mat.mut_squash();
    }
}

impl TryFrom<&MooreBox<Arb, AcbVec, AcbMat>> for MooreBox<Rif, BoxVec<Cif>, BoxMat<Cif>> {
    type Error = anyhow::Error;

    fn try_from(value: &MooreBox<Arb, AcbVec, AcbMat>) -> Result<Self, Self::Error> {
        let params = value.params.try_get_included_reckless_interval()?;
        let pt = ProjectivePoint {
            coord: value
                .pt
                .coord
                .into_iter()
                .map(|c| c.try_get_included_reckless_interval())
                .collect::<anyhow::Result<Vec<Cif>>>()?
                .into(),
            chart: value.pt.chart,
        };
        let mut mat = BoxMat::zeros(value.mat.ncols(), value.mat.nrows());
        for i in 0..mat.nrows() {
            for j in 0..mat.ncols() {
                mat[(i, j)] = value.mat[(i, j)].try_get_included_reckless_interval()?;
            }
        }

        Ok(MooreBox {
            params,
            pt,
            mat,
            rad: value.rad,
        })
    }
}

impl From<&MooreBox<Rif, BoxVec<Cif>, BoxMat<Cif>>> for MooreBox<Arb, AcbVec, AcbMat> {
    fn from(value: &MooreBox<Rif, BoxVec<Cif>, BoxMat<Cif>>) -> Self {
        value.params.squash();
        value.pt.coord.squash();
        value.mat.squash();

        let params = value.params.get_included_arb();

        let pt = ProjectivePoint {
            coord: value
                .pt
                .coord
                .0
                .into_iter()
                .map(|c| c.get_included_acb())
                .collect::<Vec<Acb>>()
                .into(),
            chart: value.pt.chart,
        };

        let mut mat = AcbMat::zeros(value.mat.ncols(), value.mat.nrows());
        for i in 0..mat.nrows() {
            for j in 0..mat.ncols() {
                mat[(i, j)] = value.mat[(i, j)].get_included_acb();
            }
        }

        MooreBox {
            params,
            pt,
            mat,
            rad: value.rad,
        }
    }
}
