// Copyright 2024 Pierre Lairez
// SPDX-License-Identifier: GPL-3.0-or-later

pub mod arith;

use anyhow::bail;
use nalgebra::Complex;
use num_traits::Zero;

use crate::{flint::Fmpq, reckless::Rif, traits::num::RefDiv};
use std::arch::x86_64::*;

#[derive(Debug, Copy, Clone)]
#[repr(C)]
pub struct Interval(pub __m256d);

impl std::fmt::Display for Interval {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        write!(f, "{} + {}*im", self.re(), self.im())
    }
}

impl std::hash::Hash for Interval {
    fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
        self.re().hash(state);
        self.im().hash(state);
    }
}

impl PartialEq for Interval {
    fn eq(&self, other: &Self) -> bool {
        self.re() == other.re() && self.im() == other.im()
    }
}

impl Eq for Interval {}

impl Interval {
    pub const IM: Self = Self::new(Rif::ZERO, Rif::ONE);
    pub const ZERO: Self = Self::new(Rif::ZERO, Rif::ZERO);
    pub const ONE: Self = Self::new(Rif::ONE, Rif::ZERO);

    fn rep(self) -> [Rif; 2] {
        unsafe { std::mem::transmute(self) }
    }

    pub fn re(self) -> Rif {
        self.rep()[0]
    }

    pub fn im(self) -> Rif {
        self.rep()[1]
    }

    pub const fn new(re: Rif, im: Rif) -> Self {
        unsafe { std::mem::transmute([re, im]) }
    }

    pub fn conj(&self) -> Self {
        Self::new(self.re(), -self.im())
    }

    pub fn abs2(&self) -> Rif {
        self.re() * self.re() + self.im() * self.im()
    }

    pub fn mag(self) -> f64 {
        self.re().mag().max(self.im().mag())
    }

    pub fn wid(self) -> f64 {
        self.re().wid().max(self.im().wid())
    }

    pub fn mid(self) -> nalgebra::Complex<f64> {
        nalgebra::Complex {
            re: self.re().mid(),
            im: self.im().mid(),
        }
    }

    pub fn squash(self) -> Self {
        Self::new(self.re().squash(), self.im().squash())
    }

    pub fn is_common(self) -> bool {
        let inf = Self::new(Rif::ENTIRE, Rif::ENTIRE);
        unsafe {
            let c = _mm256_cmp_pd(self.0, inf.0, _CMP_LT_OQ);
            _mm256_movemask_pd(c) == 15
        }
    }
}

impl From<Rif> for Interval {
    fn from(t: Rif) -> Self {
        Self::new(t, Rif::ZERO)
    }
}

impl From<i64> for Interval {
    fn from(o: i64) -> Self {
        Rif::new(o as f64, o as f64).into()
    }
}

impl Interval {
    pub fn matrix_mag<
        R: nalgebra::Dim,
        C: nalgebra::Dim,
        S: nalgebra::RawStorage<Interval, R, C>,
    >(
        mat: &nalgebra::Matrix<Interval, R, C, S>,
    ) -> f64 {
        let mut m: f64 = 0.0;
        for c in mat {
            m = m.max(c.mag());
        }
        m
    }

    pub fn matrix_wid<
        R: nalgebra::Dim,
        C: nalgebra::Dim,
        S: nalgebra::RawStorage<Interval, R, C>,
    >(
        mat: &nalgebra::Matrix<Interval, R, C, S>,
    ) -> f64 {
        let mut m: f64 = 0.0;
        for c in mat {
            m = m.max(c.wid());
        }
        m
    }
}

impl RefDiv<Interval> for Interval {
    type Output = Interval;

    fn ref_div(&self, res: &mut Self::Output, rhs: &Interval) {
        *res = *self / *rhs
    }
}

impl<'a> From<&'a Rif> for Interval {
    fn from(value: &'a Rif) -> Self {
        Self::new(*value, Rif::ZERO)
    }
}

impl From<&Complex<Fmpq>> for Interval {
    fn from(value: &Complex<Fmpq>) -> Self {
        Self::new((&value.re).into(), (&value.im).into())
    }
}

impl TryFrom<&Interval> for Rif {
    type Error = anyhow::Error;

    fn try_from(value: &Interval) -> Result<Self, Self::Error> {
        if value.im().is_zero() {
            Ok(value.re())
        } else {
            bail!("Cannot convert complex interval to real interval with non-zero real part")
        }
    }
}
