// Copyright 2024 Pierre Lairez
// SPDX-License-Identifier: GPL-3.0-or-later

use crate::{
    reckless::{complex::Interval, with_rounding_up},
    traits::num::{MulWithRop, RefNeg},
};
use num_traits::{Inv, One};
use std::{
    arch::x86_64::*,
    ops::{Add, AddAssign, Div, DivAssign, Mul, MulAssign, Neg, Sub, SubAssign},
};

type RealInterval = crate::reckless::real::Interval;

pub fn _add(x: Interval, y: Interval) -> Interval {
    Interval(unsafe { _mm256_add_pd(x.0, y.0) })
}

// see Lambov, B. (2008). Interval Arithmetic Using SSE-2. In P. Hertling, C. M.
// Hoffmann, W. Luther, & N. Revol (Eds.), Reliable Implementation of Real
// Number Algorithms: Theory and Practice (pp. 102–113). Springer.
// https://doi.org/10.1007/978-3-540-85521-7_6
fn ivchoice256(mut a: __m256d, b: __m256d) -> __m256d {
    unsafe {
        // a = [a0, a1, a2, a3]
        // b = [b0, b1, b2, b3]
        //
        // output
        // [c0, c1, c2, c3]
        // where c0 = b0 >= 0 ? a0 : -a1
        // and c1 = b1 >= 0 ? a0 : -a1
        // where c2 = b2 >= 0 ? a2 : -a3
        // and c2 = b3 >= 0 ? a2 : -a3

        // a = [a0, -a1]
        // _mm_set_pd has the low part as last argument
        a = _mm256_xor_pd(a, _mm256_set_pd(-0.0, 0.0, -0.0, 0.0));

        // cmp[1] = sign bit of b0
        // cmp[65] = sign bit of b1
        let cmp = _mm256_srli_epi64(_mm256_castpd_si256(b), 62); // AVX2
                                                                 //let cmp = _mm256_castpd_si256(_mm256_cmp_pd(b, _mm256_set_pd(0.0, 0.0, 0.0, 0.0), _CMP_LT_OQ));
        _mm256_permutevar_pd(a, cmp)
    }
}

pub fn _mul_componentwise(x: Interval, y: Interval) -> Interval {
    let a = x.0;
    let b = y.0;

    unsafe {
        #[cfg(debug_assertions)]
        {
            #[allow(deprecated)]
            let rnd = _MM_GET_ROUNDING_MODE();
            assert!(rnd == _MM_ROUND_UP, "Wrong rounding mode");
        }

        let arev = _mm256_permute_pd(a, 5);
        let brev = _mm256_permute_pd(b, 5);

        let mut u = _mm256_mul_pd(ivchoice256(a, brev), brev);
        let mut v = _mm256_mul_pd(ivchoice256(arev, b), b);

        let cu = _mm256_cmp_pd(u, u, _CMP_EQ_OQ);
        let cv = _mm256_cmp_pd(v, v, _CMP_EQ_OQ);

        // if _mm_movemask_pd(_mm_and_pd(cu, cv)) != 3 {
        // NaNs may come from a multiplication ∞ × 0, which should be 0
        // in the context of interval arithmetic, or from the input.

        // replace NaNs with zeros
        u = _mm256_and_pd(cu, u);
        v = _mm256_and_pd(cv, v);

        let w = _mm256_max_pd(u, v);

        Interval(w)
    }
}

pub fn _mul_interval(x: Interval, y: RealInterval) -> Interval {
    let yy = Interval::new(y, y);
    _mul_componentwise(x, yy)
}

#[inline(always)]
pub fn _mul(x: Interval, y: Interval) -> Interval {
    let aa = Interval::new(x.re(), x.re());
    let bb = Interval::new(x.im(), x.im());
    let mdc = Interval::new(-y.im(), y.re());

    _add(_mul_componentwise(aa, y), _mul_componentwise(bb, mdc))
}

#[inline(always)]
pub fn _mul_componentwise_bounded(x: Interval, y: Interval) -> Interval {
    let a = x.0;
    let b = y.0;

    unsafe {
        #[cfg(debug_assertions)]
        {
            #[allow(deprecated)]
            let rnd = _MM_GET_ROUNDING_MODE();
            assert!(rnd == _MM_ROUND_UP, "Wrong rounding mode");
        }

        let arev = _mm256_permute_pd(a, 5);
        let brev = _mm256_permute_pd(b, 5);

        let u = _mm256_mul_pd(ivchoice256(a, brev), brev);
        let v = _mm256_mul_pd(ivchoice256(arev, b), b);

        let w = _mm256_max_pd(u, v);
        Interval(w)
    }
}

//#[inline(always)]
pub fn _mul_bounded(x: Interval, y: Interval) -> Interval {
    let aa = Interval::new(x.re(), x.re());
    let bb = Interval::new(x.im(), x.im());
    let mdc = Interval::new(-y.im(), y.re());

    _add(_mul_componentwise_bounded(aa, y), _mul_componentwise_bounded(bb, mdc))
}

// Arithmetic operations
impl Mul<RealInterval> for Interval {
    type Output = Self;
    fn mul(self, other: RealInterval) -> Self {
        with_rounding_up(|| Self::new(self.re() * other, self.im() * other))
    }
}

impl<'a> Mul<&'a RealInterval> for Interval {
    type Output = Self;
    fn mul(self, other: &'a RealInterval) -> Self {
        self * *other
    }
}

impl MulAssign<RealInterval> for Interval {
    fn mul_assign(&mut self, other: RealInterval) {
        *self = *self * other;
    }
}

impl MulAssign<&RealInterval> for Interval {
    fn mul_assign(&mut self, other: &RealInterval) {
        *self = *self * other;
    }
}

impl num_traits::identities::Zero for Interval {
    fn zero() -> Self {
        Self::ZERO
    }
    fn is_zero(&self) -> bool {
        *self == Self::ZERO
    }
    fn set_zero(&mut self) {
        *self = Self::ZERO;
    }
}

impl num_traits::identities::One for Interval {
    fn one() -> Self {
        Self::ONE
    }
    fn is_one(&self) -> bool {
        *self == Self::ONE
    }
    fn set_one(&mut self) {
        *self = Self::ONE;
    }
}

//. Addition
//.. lhs: value, rhs: value
impl Add for Interval {
    type Output = Self;
    fn add(self, other: Self) -> Self {
        with_rounding_up(|| _add(self, other))
    }
}

//.. lhs: value, rhs: ref
impl<'a> Add<&'a Interval> for Interval {
    type Output = Interval;

    fn add(self, rhs: &'a Interval) -> Self::Output {
        self + *rhs
    }
}

//.. lhs: ref, rhs: value
impl<'a> Add<Interval> for &'a Interval {
    type Output = Interval;

    fn add(self, rhs: Interval) -> Self::Output {
        *self + rhs
    }
}

//.. lhs: ref, rhs: ref
impl<'a, 'b> Add<&'a Interval> for &'b Interval {
    type Output = Interval;

    fn add(self, rhs: &'a Interval) -> Self::Output {
        *self + *rhs
    }
}

//.. rhs: value
impl AddAssign for Interval {
    fn add_assign(&mut self, other: Self) {
        *self = *self + other;
    }
}

//.. rhs: ref
impl<'a> AddAssign<&'a Interval> for Interval {
    fn add_assign(&mut self, rhs: &'a Interval) {
        *self += *rhs
    }
}

//. Substraction
impl Neg for Interval {
    type Output = Interval;
    fn neg(self) -> Self::Output {
        Interval(unsafe { _mm256_permute_pd(self.0, 5) })
    }
}

//.. lhs: value, rhs: value
impl Sub for Interval {
    type Output = Self;
    fn sub(self, other: Self) -> Self {
        self + (-other)
    }
}

//.. lhs: value, rhs: ref
impl<'a> Sub<&'a Interval> for Interval {
    type Output = Interval;

    fn sub(self, rhs: &'a Interval) -> Self::Output {
        self - *rhs
    }
}

//.. lhs: ref, rhs: value
impl<'a> Sub<Interval> for &'a Interval {
    type Output = Interval;

    fn sub(self, rhs: Interval) -> Self::Output {
        *self - rhs
    }
}

//.. lhs: ref, rhs: ref
impl<'a, 'b> Sub<&'a Interval> for &'b Interval {
    type Output = Interval;

    fn sub(self, rhs: &'a Interval) -> Self::Output {
        *self - *rhs
    }
}

//.. rhs: value
impl SubAssign for Interval {
    fn sub_assign(&mut self, other: Self) {
        *self = *self - other;
    }
}

//.. rhs: ref
impl<'a> SubAssign<&'a Interval> for Interval {
    fn sub_assign(&mut self, rhs: &'a Interval) {
        *self -= *rhs
    }
}

//. Multiplication
//.. lhs: value, rhs: value
impl Mul for Interval {
    type Output = Self;
    fn mul(self, other: Self) -> Self {
        with_rounding_up(|| _mul(self, other))
        // Self::new(
        //     self.re() * other.re() - self.im() * other.im(),
        //     self.im() * other.re() + self.re() * other.im(),
        // )
    }
}

//.. lhs: value, rhs: ref
impl<'a> Mul<&'a Interval> for Interval {
    type Output = Interval;

    fn mul(self, rhs: &'a Interval) -> Self::Output {
        self * *rhs
    }
}

//.. lhs: ref, rhs: value
impl<'a> Mul<Interval> for &'a Interval {
    type Output = Interval;

    fn mul(self, rhs: Interval) -> Self::Output {
        *self * rhs
    }
}

//.. lhs: ref, rhs: ref
impl<'a, 'b> Mul<&'a Interval> for &'b Interval {
    type Output = Interval;

    fn mul(self, rhs: &'a Interval) -> Self::Output {
        *self * *rhs
    }
}

//.. rhs: value
impl MulAssign for Interval {
    fn mul_assign(&mut self, other: Interval) {
        *self = *self * other;
    }
}

//.. rhs: ref
impl<'a> MulAssign<&'a Interval> for Interval {
    fn mul_assign(&mut self, rhs: &'a Interval) {
        *self *= *rhs
    }
}

//. Division
impl Inv for Interval {
    type Output = Self;

    fn inv(self) -> Self::Output {
        Interval::one() / self
    }
}

//.. lhs: value, rhs: value
impl Div for Interval {
    type Output = Self;
    fn div(self, other: Self) -> Self {
        self * other.conj() * other.abs2().recip()
    }
}

//.. lhs: value, rhs: ref
impl<'a> Div<&'a Interval> for Interval {
    type Output = Interval;

    fn div(self, rhs: &'a Interval) -> Self::Output {
        self / *rhs
    }
}

//.. lhs: ref, rhs: value
impl<'a> Div<Interval> for &'a Interval {
    type Output = Interval;

    fn div(self, rhs: Interval) -> Self::Output {
        *self / rhs
    }
}

//.. lhs: ref, rhs: ref
impl<'a, 'b> Div<&'a Interval> for &'b Interval {
    type Output = Interval;

    fn div(self, rhs: &'a Interval) -> Self::Output {
        *self / *rhs
    }
}

//.. lhs: value
impl DivAssign for Interval {
    fn div_assign(&mut self, rhs: Self) {
        *self = *self / rhs
    }
}

//.. lhs: ref
impl<'a> DivAssign<&'a Interval> for Interval {
    fn div_assign(&mut self, rhs: &'a Interval) {
        *self /= *rhs
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn add() {
        with_rounding_up(|| {
            assert_eq!(
                _add(
                    Interval::new(RealInterval::new(2.0, 3.0), RealInterval::new(4.0, 6.0)),
                    Interval::new(RealInterval::new(10.0, 20.0), RealInterval::new(40.0, 60.0))
                ),
                Interval::new(RealInterval::new(12.0, 23.0), RealInterval::new(44.0, 66.0))
            )
        })
    }

    #[test]
    fn mul() {
        with_rounding_up(|| {
            assert_eq!(
                _mul(
                    Interval::new(RealInterval::new(2.0, 3.0), RealInterval::new(4.0, 6.0)),
                    Interval::new(RealInterval::new(10.0, 20.0), RealInterval::new(40.0, 60.0))
                ),
                Interval::new(RealInterval::new(-340.0, -100.0), RealInterval::new(120.0, 300.0))
            );

            assert_eq!(_mul_interval(Interval::ZERO, RealInterval::ENTIRE), Interval::ZERO);
        })
    }
}

impl RefNeg for Interval {
    fn ref_neg(&mut self) {
        *self = -*self;
    }
}

impl MulWithRop<Interval> for Interval {
    type Output = Interval;

    fn mul_with_rop(&self, rhs: &Interval, res: &mut Self::Output) {
        *res = *self * *rhs;
    }
}
