use crate::{
    flint::Arb,
    reckless::{Cif, Rif},
};

pub trait Interval: Clone {
    // returns a singleton included in self (moreover the singleton should lie around the center of self)
    fn squash(&self) -> Self;
    // in place squash
    fn mut_squash(&mut self);
    // returns an upperbound for the magnitude of self as a f64
    fn mag(&self) -> f64;
    // returns a lower bound for the magnitude of self as a f64
    fn mag_lower(&self) -> f64;
    // returns an upperbound for the diameter of self as a f64
    fn diam(&self) -> f64;
    // returns a box of center 0 and of radius rad (or at least a box containing the theoritical box of center 0 and radius r).
    fn square(rad: f64) -> Self;
    // returns a box (tightly) containing self and other
    fn union(&self, other: &Self) -> Self;
}

impl Interval for Cif {
    fn diam(&self) -> f64 {
        (*self).wid()
    }

    fn mag(&self) -> f64 {
        (*self).mag()
    }

    fn mag_lower(&self) -> f64 {
        self.mag()
    }

    fn squash(&self) -> Self {
        (*self).squash()
    }

    fn mut_squash(&mut self) {
        *self = self.squash()
    }

    fn square(r: f64) -> Self {
        Self::new(Rif::new(-r, r), Rif::new(-r, r))
    }

    fn union(&self, other: &Self) -> Self {
        Self::new(self.re().union(&other.re()), self.im().union(&other.im()))
    }
}

impl Interval for Rif {
    fn diam(&self) -> f64 {
        (*self).wid()
    }

    fn mag(&self) -> f64 {
        (*self).mag()
    }

    fn mag_lower(&self) -> f64 {
        self.mag()
    }

    fn squash(&self) -> Self {
        (*self).squash()
    }

    fn mut_squash(&mut self) {
        *self = self.squash()
    }

    fn square(r: f64) -> Self {
        Self::new(-r, r)
    }

    fn union(&self, other: &Self) -> Self {
        Self::new(self._inf().min(other._inf()), self._sup().max(other._sup()))
    }
}

pub trait RealInterval: Interval {
    // returns an (singleton) interval sup such that self <= sup
    fn sup(&self) -> Self;
    // returns an (singleton) interval inf such that self >= inf
    fn inf(&self) -> Self;
    // returns an interval (tightly) containing the intersection of self and other
    fn intersection(&self, other: &Self) -> Self;
    // returns an interval (tightly) containing [a, b] where a is the min of inf and b is the max of sup. It is assumed that a <= b. Usually, inf and sup should be singletons (so a resp. b and inf resp. sup should coincide)
    fn from_inf_sup(inf: Self, sup: Self) -> Self;
    // returns [-∞, ∞]
    fn entire() -> Self;
    // returns self ⊆ other
    fn subset(&self, other: &Self) -> bool;
}

impl RealInterval for Rif {
    fn intersection(&self, other: &Self) -> Self {
        (*self).intersection(*other)
    }

    fn sup(&self) -> Self {
        (*self)._sup().into()
    }

    fn inf(&self) -> Self {
        (*self)._inf().into()
    }

    fn from_inf_sup(inf: Self, sup: Self) -> Self {
        Self::new(inf._inf(), sup._sup())
    }

    fn entire() -> Self {
        Self::ENTIRE
    }

    fn subset(&self, other: &Self) -> bool {
        (*self).subset(*other)
    }
}

// It is tempting to implement PartialOrd for intervals, but is a bad idea as:
// - if one overloads default implementations of comparisons in PartialOrd to match with interval comparisons, leading to logical inconsistencies. This is dangerous as one could use methods that require PartialOrd that assume the logical consistency of the different comparison operators.
// - if one implements partial_cmp only, the default implementations of the different comparisons will not match interval comparisons.
// A possibility would be to implement the partial order to compare only singletons, but this forces equality to be tested only on singletons to ensure logical consistency (though it seems natural to have the set equality implemented in PartialEq). This idea is quite misleading as x ≤ y does not correspond to the interval leq, so one could write x ≤ y thinking in interval arithmetic terms, but the test would fail if x or y are not singletons. For those reasons, it seems advised not to implement PartialOrd for intervals, and instead write a new trait.
pub trait IntervalComparisons: RealInterval {
    // If op is one of the operators =, ≠, ≤, <, ≥, >, then its associated interval comparison should return true on X, Y iff for all x ∈ X, for all y ∈ Y, x op y
    fn interval_eq(&self, other: &Self) -> bool;
    fn interval_ne(&self, other: &Self) -> bool;
    fn interval_le(&self, other: &Self) -> bool;
    fn interval_lt(&self, other: &Self) -> bool;
    fn interval_ge(&self, other: &Self) -> bool;
    fn interval_gt(&self, other: &Self) -> bool;
}

impl IntervalComparisons for Rif {
    fn interval_eq(&self, other: &Self) -> bool {
        self.wid() == 0.0 && other.wid() == 0.0 && self.lo() == other.lo()
    }

    fn interval_ne(&self, other: &Self) -> bool {
        self.disjoint(*other)
    }

    fn interval_ge(&self, other: &Self) -> bool {
        other.precedes(*self)
    }

    fn interval_gt(&self, other: &Self) -> bool {
        other.strict_precedes(*self)
    }

    fn interval_le(&self, other: &Self) -> bool {
        self.precedes(*other)
    }

    fn interval_lt(&self, other: &Self) -> bool {
        self.strict_precedes(*other)
    }
}

// To convert intervals X s.t. ∀x ∈ X, P(x) for some property P
pub trait IntervalIncludedFrom<T>
where
    Self: Sized,
{
    fn interval_included_from(other: T) -> anyhow::Result<Self>;
}

// To convert intervals X s.t. ∃x ∈ X, P(x) for some property P
pub trait IntervalCoveringFrom<T>
where
    Self: Sized,
{
    fn interval_covering_from(other: T) -> anyhow::Result<Self>;
}

impl IntervalIncludedFrom<&Rif> for Arb {
    fn interval_included_from(other: &Rif) -> anyhow::Result<Self> {
        Ok(other.get_included_arb())
    }
}

impl IntervalIncludedFrom<&Arb> for Rif {
    fn interval_included_from(other: &Arb) -> anyhow::Result<Self> {
        other.try_get_included_reckless_interval()
    }
}

impl IntervalCoveringFrom<&Rif> for Arb {
    fn interval_covering_from(other: &Rif) -> anyhow::Result<Self> {
        unimplemented!()
    }
}

impl IntervalCoveringFrom<&Arb> for Rif {
    fn interval_covering_from(other: &Arb) -> anyhow::Result<Self> {
        Ok(other.get_covering_reckless_interval())
    }
}
