use crate::{
    flint::{Acb, Fmpq},
    foa::FaATrait,
    precision_ctx::get_prec,
    prog::{AcbProg, ExactProg, Program, RecklessProg},
    projective::ParametricSystem,
};
use nalgebra::Complex;
use std::sync::{Arc, RwLock};

pub type RecklessSys = ParametricSystem<RecklessProg>;

impl From<ParametricSystem<ExactProg>> for RecklessSys {
    fn from(value: ParametricSystem<ExactProg>) -> Self {
        ParametricSystem {
            nparams: value.nparams,
            dim: value.dim,
            chart_switch_threshold: value.chart_switch_threshold,
            f: value.f.into_iter().map(|c| RecklessProg(from_prog(c))).collect(),
            df: value.df.into_iter().map(|c| RecklessProg(from_prog(c))).collect(),
            fdot: value.fdot.into_iter().map(|c| RecklessProg(from_prog(c))).collect(),
        }
    }
}

pub type AcbSys = ParametricSystem<AcbProg>;

impl From<ParametricSystem<ExactProg>> for AcbSys {
    fn from(value: ParametricSystem<ExactProg>) -> Self {
        let prog_convert = |prog: ExactProg| {
            let acb_scalars: Vec<Acb> = prog.scalars.read().unwrap().iter().map(|scalar| Acb::from(scalar)).collect();
            AcbProg {
                program: Program {
                    scalars: RwLock::new(Arc::new(acb_scalars)),
                    instructions: prog.instructions,
                    vars: prog.vars,
                    outnodes: prog.outnodes,
                    bufferlen: prog.bufferlen,
                },
                exact_coeffs: (**prog.scalars.read().unwrap()).clone(), //TODO: get rid of this clone
                prec: RwLock::new(get_prec()),
            }
        };
        ParametricSystem {
            nparams: value.nparams,
            dim: value.dim,
            chart_switch_threshold: value.chart_switch_threshold,
            f: value.f.into_iter().map(|c| prog_convert(c)).collect(),
            df: value.df.into_iter().map(|c| prog_convert(c)).collect(),
            fdot: value.fdot.into_iter().map(|c| prog_convert(c)).collect(),
        }
    }
}

pub struct AcbRecklessSys {
    pub acb_sys: AcbSys,
    pub reckless_sys: RecklessSys,
}

impl FaATrait for AcbRecklessSys {
    type A = AcbSys;
    type F = RecklessSys;

    fn get_accurate(&self) -> &Self::A {
        &self.acb_sys
    }

    fn get_fast(&self) -> &Self::F {
        &self.reckless_sys
    }
}

impl From<ParametricSystem<ExactProg>> for AcbRecklessSys {
    fn from(value: ParametricSystem<ExactProg>) -> Self {
        let ref_prog_convert = |prog: &ExactProg| {
            let acb_scalars: Vec<Acb> = prog.scalars.read().unwrap().iter().map(|scalar| Acb::from(scalar)).collect();
            AcbProg {
                program: Program {
                    scalars: RwLock::new(Arc::new(acb_scalars)),
                    instructions: prog.instructions.clone(),
                    vars: prog.vars.clone(),
                    outnodes: prog.outnodes.clone(),
                    bufferlen: prog.bufferlen,
                },
                exact_coeffs: (**prog.scalars.read().unwrap()).clone(),
                prec: RwLock::new(get_prec()),
            }
        };
        AcbRecklessSys {
            acb_sys: ParametricSystem {
                nparams: value.nparams,
                dim: value.dim,
                chart_switch_threshold: value.chart_switch_threshold,
                f: value.f.iter().map(|c| ref_prog_convert(c)).collect(),
                df: value.df.iter().map(|c| ref_prog_convert(c)).collect(),
                fdot: value.fdot.iter().map(|c| ref_prog_convert(c)).collect(),
            },
            reckless_sys: value.into(),
        }
    }
}

fn from_prog<C>(prog: ExactProg) -> Program<C>
where
    for<'r> C: From<&'r Complex<Fmpq>>,
{
    let exact_scalars = prog.get_scalars();
    Program {
        scalars: RwLock::new(Arc::new(
            exact_scalars.iter().map(|exact_scalar| exact_scalar.into()).collect(),
        )),
        instructions: prog.instructions,
        vars: prog.vars,
        outnodes: prog.outnodes,
        bufferlen: prog.bufferlen,
    }
}
