use super::{
    interval::IntervalComparisons,
    system::{AcbRecklessSys, AcbSys, RecklessSys},
    vector::Vector,
    Interval,
};
use crate::{
    box_mat::BoxMat,
    box_vec::BoxVec,
    flint::{Acb, AcbMat, AcbTm, AcbVec, AcbVecTm, Arb},
    foa::FoA,
    precision_ctx::{AdaptivePrecCtx, FixedPrecCtx, PrecCtx},
    projective::{MooreBox, ProjectivePoint},
    reckless::{Cif, Rif},
    reckless_extra::RecklessTm5,
    taylor_model::tm::RecklessTm,
};
use anyhow::bail;
use flint_bindings::arb_set_d;
use num_traits::Zero;
use one_step::{ArbOneStep, ArbRecklessOneStep, GMixed, OneStep, RecklessOneStep, RecklessOneStep5};
use std::{fmt::{Debug, Display}, marker::PhantomData};

mod distort;
mod extend;
pub mod find_moore_box;
mod one_step;
mod predictor;
mod refine;
// pub mod refine_iter;
pub mod to_chart;

pub trait Track {
    type Sys;
    type R;
    type VecC;
    type MatC;
    type VecTm;

    fn track(
        exact_sys: &Self::Sys,
        mbox: MooreBox<Self::R, Self::VecC, Self::MatC>,
    ) -> anyhow::Result<(
        MooreBox<Self::R, Self::VecC, Self::MatC>,
        Vec<(Self::R, ProjectivePoint<Self::VecTm>, i64, f64)>,
    )>;
}

pub struct GTrack<Ctx, Sys, R, C, VecC, MatC, Tm, VecTm, GOneStep> {
    _phantom: PhantomData<(Ctx, Sys, R, C, VecC, MatC, Tm, VecTm, GOneStep)>,
}

impl<Ctx, Sys, R, C, VecC, MatC, Tm, VecTm, GOneStep> Track for GTrack<Ctx, Sys, R, C, VecC, MatC, Tm, VecTm, GOneStep>
where
    VecC: Vector<T = C> + Interval + Debug,
    MatC: Debug,
    R: Debug + Display,
    C: Zero + Debug,
    for<'l> &'l C: TryInto<R, Error = anyhow::Error>,
    R: From<f64> + IntervalComparisons,
    Ctx: PrecCtx,
    GOneStep: OneStep<Sys = Sys, R = R, VecC = VecC, MatC = MatC, Tm = Tm, VecTm = VecTm>,
{
    type R = R;
    type VecC = VecC;
    type MatC = MatC;
    type VecTm = VecTm;
    type Sys = Sys;

    //
    #[inline(never)]
    fn track(
        sys: &Self::Sys,
        mut mbox: MooreBox<Self::R, Self::VecC, Self::MatC>,
    ) -> anyhow::Result<(
        MooreBox<Self::R, Self::VecC, Self::MatC>,
        Vec<(Self::R, ProjectivePoint<Self::VecTm>, i64, f64)>,
    )> {
        let mut step_size = 1.0;
        let mut ret = Vec::new();

        // This information is useful for the Hermite predictor
        let mut previous_point = None;
        let mut previous_speed = None;

        #[cfg(debug_assertions)]
        let mut cpt = 1;

        while mbox.params.interval_lt(&1.0.into()) {
            #[cfg(debug_assertions)]
            {
                if cpt % 1 == 0 {
                    // dbg!(Ctx::get_prec());
                    // dbg!(&mbox.pt.coord.mag());
                    // dbg!(step_size);
                    // dbg!(cpt);
                    // println!("{}", &mbox.params);
                    // dbg!(&mbox.rad);
                    // dbg!(&mbox.pt.coord);
                    // dbg!(&mbox.pt.chart);
                }
                cpt += 1;
            }

            let p = Ctx::get_prec();
            if p > 53 {
                let _ = Ctx::set_prec(p - 1);
            }
            let save_t = mbox.params.clone();
            let s;
            let tm;
            (mbox, s, tm) = GOneStep::one_step(sys, mbox, step_size, &mut previous_point, &mut previous_speed)?;
            step_size = s;
            ret.push((save_t, tm, Ctx::get_prec(), s));
        }

        #[cfg(debug_assertions)]
        {
            dbg!(step_size);
            dbg!(Ctx::get_prec());
            dbg!(&mbox.pt.coord);
            dbg!(&mbox.pt.chart);
        }
        Ok((mbox, ret))
    }
}

pub type RecklessTrack5 =
    GTrack<FixedPrecCtx, RecklessSys, Rif, Cif, BoxVec<Cif>, BoxMat<Cif>, RecklessTm5, BoxVec<RecklessTm5>, RecklessOneStep5>;

pub type RecklessTrack<const N: usize> = GTrack<
    FixedPrecCtx,
    RecklessSys,
    Rif,
    Cif,
    BoxVec<Cif>,
    BoxMat<Cif>,
    RecklessTm<N>,
    BoxVec<RecklessTm<N>>,
    RecklessOneStep<N>,
>;

pub type ArbTrack<const N: usize> =
    GTrack<AdaptivePrecCtx, AcbSys, Arb, Acb, AcbVec, AcbMat, AcbTm<N>, AcbVecTm<N>, ArbOneStep<N>>;

pub type ArbReckless5Track =
    GTrack<AdaptivePrecCtx, AcbRecklessSys, Arb, Acb, AcbVec, AcbMat, AcbTm<5>, AcbVecTm<5>, ArbRecklessOneStep>;

pub type ArbReckless5Track2 = GTrack<
    AdaptivePrecCtx,
    AcbRecklessSys,
    FoA<Rif, Arb>,
    FoA<Cif, Acb>,
    FoA<BoxVec<Cif>, AcbVec>,
    FoA<BoxMat<Cif>, AcbMat>,
    FoA<RecklessTm5, AcbTm<5>>,
    FoA<BoxVec<RecklessTm5>, AcbVecTm<5>>,
    GMixed<AcbRecklessSys, RecklessOneStep5, ArbOneStep<5>>,
>;

impl Rif {
    pub fn get_included_arb(&self) -> Arb {
        let mut res = Arb::default();
        unsafe {
            arb_set_d(&mut res.0, self.mid());
        }
        res
    }
}

impl Cif {
    pub fn get_included_acb(&self) -> Acb {
        Acb::from_re_im(&self.re().get_included_arb(), &self.im().get_included_arb())
    }
}

#[cfg(test)]
mod marker {
    use super::{ArbReckless5Track, ArbReckless5Track2, ArbTrack, RecklessTrack, RecklessTrack5, Track};

    fn _marker<T: Track>() {}

    #[test]
    fn marker() {
        _marker::<RecklessTrack5>();
        _marker::<RecklessTrack<20>>();
        _marker::<ArbTrack<20>>();
        _marker::<ArbReckless5Track>();
        _marker::<ArbReckless5Track2>();
    }
}
