use crate::{
    box_mat::BoxMat,
    box_vec::BoxVec,
    flint::AcbMat,
    prog::{eval_map, AcbProg, Eval, FlintProg, ProgramTrait, RecklessProg},
    projective::{ParametricSystem, ProjectivePoint},
    reckless::{with_rounding_up, Cif},
    reckless_extra::{RecklessTm5, UnsafeCif, UnsafeRecklessTm5},
    taylor_model::tm::RecklessTm,
    traits::{matrix::Matrix, num::MulWithRop, vector::Vector, Interval},
};
use num_traits::{One, Zero};
use std::{
    marker::PhantomData,
    ops::{Add, AddAssign, MulAssign, Neg},
};

pub trait Distort<VecS, MatS> {
    type Sys;
    type S;
    type MatC;

    fn distort(sys: &Self::Sys, params: &Self::S, pt: &ProjectivePoint<VecS>, mat: &Self::MatC, rad: f64) -> VecS;
}

pub trait _Distort<VecS, MatS> {
    type MatC;

    fn _distort(df: &MatS, res: &mut VecS, mat: &Self::MatC);
}

pub struct _GDistort<MatC> {
    phantom: PhantomData<MatC>,
}

impl<MatC, VecS, MatS> _Distort<VecS, MatS> for _GDistort<MatC>
where
    MatC: Matrix,
    MatC::Coeffs: One + Neg<Output = MatC::Coeffs> + Interval,
    VecS: Vector<T = MatS::Coeffs>,
    MatS: Matrix,
    MatS::Coeffs: Zero
        + MulWithRop<MatC::Coeffs, Output = MatS::Coeffs>
        + for<'a> AddAssign<&'a MatS::Coeffs>
        + for<'a> AddAssign<&'a MatC::Coeffs>
        + for<'a> MulAssign<&'a MatC::Coeffs>,
{
    type MatC = MatC;

    #[inline(always)]
    fn _distort(df: &MatS, res: &mut VecS, mat: &Self::MatC) {
        let n = df.nrows();
        let mut p = MatS::Coeffs::zero();
        let mut m = MatS::Coeffs::zero();

        for i in 0..n {
            for k in 0..n {
                m.set_zero();
                for j in 0..n {
                    df[(j, k)].mul_with_rop(&mat[(i, j)], &mut p);
                    m += &p;
                }
                if i == k {
                    m += &(-MatC::Coeffs::one());
                }
                m *= &MatC::Coeffs::square(1.0);
                res[i] += &m;
            }
        }
    }
}

pub struct GDistort<Prog, _GDistort> {
    phantom: PhantomData<(Prog, _GDistort)>,
}

impl<Prog, _GDistort, VecS, MatS> Distort<VecS, MatS> for GDistort<Prog, _GDistort>
where
    Prog: ProgramTrait,
    for<'r1, 'r2> Prog: Eval<(&'r1 VecS::T, &'r2 VecS), Vec<VecS::T>>,
    for<'r> Prog: Eval<(), Vec<VecS::T>>,
    VecS: Vector,
    VecS::T: Zero + Clone + 'static,
    for<'l> &'l VecS::T: Add<<_GDistort::MatC as Matrix>::Coeffs, Output = VecS::T>,
    MatS: Matrix<Coeffs = VecS::T>,
    _GDistort::MatC: Matrix,
    <_GDistort::MatC as Matrix>::Coeffs: Interval,
    _GDistort: _Distort<VecS, MatS>,
{
    type Sys = ParametricSystem<Prog>;
    type S = VecS::T;
    type MatC = _GDistort::MatC;

    fn distort(sys: &Self::Sys, params: &Self::S, pt: &ProjectivePoint<VecS>, mat: &Self::MatC, rad: f64) -> VecS {
        let n = pt.coord.len();
        let df = MatS::from_vec(
            n,
            n,
            eval_map(&sys.df[pt.chart], |i| {
                if i < 1 {
                    Some(params.clone())
                } else if i - 1 < pt.coord.len() {
                    Some(&pt.coord[i - 1] + <_GDistort::MatC as Matrix>::Coeffs::square(rad))
                } else {
                    Some(VecS::T::zero())
                }
            }),
        );
        let mut res = VecS::zeros(pt.coord.len());
        _GDistort::_distort(&df, &mut res, &mat);
        res
    }
}

pub type ArbDistort = GDistort<AcbProg, _GDistort<AcbMat>>;

pub type RecklessDistort = GDistort<RecklessProg, _SafeRecklessDistort>;

pub struct _SafeRecklessDistort {}

impl _Distort<BoxVec<Cif>, BoxMat<Cif>> for _SafeRecklessDistort {
    type MatC = BoxMat<Cif>;

    fn _distort(df: &BoxMat<Cif>, res: &mut BoxVec<Cif>, mat: &Self::MatC) {
        unsafe {
            with_rounding_up(|| {
                _GDistort::<BoxMat<UnsafeCif>>::_distort(
                    &*(df as *const BoxMat<Cif> as *const BoxMat<UnsafeCif>),
                    &mut *(res as *mut BoxVec<Cif> as *mut BoxVec<UnsafeCif>),
                    &*(mat as *const BoxMat<Cif> as *const BoxMat<UnsafeCif>),
                )
            })
        }
    }
}

impl _Distort<BoxVec<RecklessTm5>, BoxMat<RecklessTm5>> for _SafeRecklessDistort {
    type MatC = BoxMat<Cif>;

    fn _distort(df: &BoxMat<RecklessTm5>, res: &mut BoxVec<RecklessTm5>, mat: &Self::MatC) {
        unsafe {
            with_rounding_up(|| {
                _GDistort::<BoxMat<UnsafeCif>>::_distort(
                    &*(df as *const BoxMat<RecklessTm5> as *const BoxMat<UnsafeRecklessTm5>),
                    &mut *(res as *mut BoxVec<RecklessTm5> as *mut BoxVec<UnsafeRecklessTm5>),
                    &*(mat as *const BoxMat<Cif> as *const BoxMat<UnsafeCif>),
                )
            })
        }
    }
}

impl<const N: usize> _Distort<BoxVec<RecklessTm<N>>, BoxMat<RecklessTm<N>>> for _SafeRecklessDistort {
    type MatC = BoxMat<Cif>;

    fn _distort(df: &BoxMat<RecklessTm<N>>, res: &mut BoxVec<RecklessTm<N>>, mat: &Self::MatC) {
        _GDistort::<BoxMat<Cif>>::_distort(df, res, mat);
    }
}

#[cfg(test)]
mod marker_tests {
    use super::{ArbDistort, Distort, RecklessDistort};
    use crate::{
        box_mat::BoxMat,
        box_vec::BoxVec,
        flint::{AcbMat, AcbVec},
        reckless::Cif,
        reckless_extra::RecklessTm5,
        taylor_model::tm::RecklessTm,
    };

    fn _marker<VecC, MatC, T: Distort<VecC, MatC>>() {}

    #[test]
    fn marker_tests() {
        _marker::<BoxVec<Cif>, BoxMat<Cif>, RecklessDistort>();
        _marker::<BoxVec<RecklessTm5>, BoxMat<RecklessTm5>, RecklessDistort>();
        _marker::<BoxVec<RecklessTm<20>>, BoxMat<RecklessTm<20>>, RecklessDistort>();
        _marker::<AcbVec, AcbMat, ArbDistort>();
    }
}
