use super::{
    extend::{ArbExtend, Extend, RecklessExtend, RecklessExtend5},
    refine::{ArbRefine, RecklessRefine, Refine},
    to_chart::{ArbToChart, RecklessToChart, ToChart},
};
use crate::{
    box_mat::BoxMat,
    box_vec::BoxVec,
    flint::{Acb, AcbMat, AcbTm, AcbVec, AcbVecTm, Arb},
    foa::{FaATrait, FoA, FoATrait},
    precision_ctx::{get_prec, AdaptivePrecCtx, FixedPrecCtx, PrecCtx},
    prog::{AcbProg, RecklessProg},
    projective::{MooreBox, ParametricSystem, ProjectivePoint},
    reckless::{Cif, Rif},
    reckless_extra::RecklessTm5,
    taylor_model::tm::RecklessTm,
    traits::{system::AcbRecklessSys, vector::Vector, Interval},
};
use anyhow::bail;
use num_traits::{One, Zero};
use std::{fmt::Debug, marker::PhantomData};

pub trait OneStep {
    type Sys;
    type R;
    type VecC;
    type MatC;
    type Tm;
    type VecTm;

    fn one_step(
        sys: &Self::Sys,
        mbox: MooreBox<Self::R, Self::VecC, Self::MatC>,
        step_size: f64,
        previous_pt: &mut Option<Self::VecC>,
        previous_speed: &mut Option<Self::VecC>,
    ) -> anyhow::Result<(MooreBox<Self::R, Self::VecC, Self::MatC>, f64, ProjectivePoint<Self::VecTm>)>;
}

pub struct GOneStep<GRefine, GExtend, GToChart, Ctx, Prog, R, C, VecC, MatC, Tm, VecTm> {
    phantom: PhantomData<(GRefine, GExtend, GToChart, Ctx, Prog, R, C, VecC, MatC, Tm, VecTm)>,
}

impl<GRefine, GExtend, GToChart, Ctx, Prog, R, C, VecC, MatC, Tm, VecTm> OneStep
    for GOneStep<GRefine, GExtend, GToChart, Ctx, Prog, R, C, VecC, MatC, Tm, VecTm>
where
    Ctx: PrecCtx,
    GRefine: Refine<Sys = ParametricSystem<Prog>, R = R, VecC = VecC, MatC = MatC>,
    GExtend: Extend<Sys = ParametricSystem<Prog>, R = R, VecC = VecC, MatC = MatC, Tm = Tm, VecTm = VecTm>,
    GToChart: ToChart<MatC = MatC, VecC = VecC, Sys = ParametricSystem<Prog>, R = R>,
    VecC: Interval + Debug + Vector + Debug,
    VecC::T: Clone + One + Interval + Zero,
    MatC: Interval,
    R: Interval,
{
    type Sys = ParametricSystem<Prog>;
    type R = R;
    type VecC = VecC;
    type MatC = MatC;
    type Tm = Tm;
    type VecTm = VecTm;

    fn one_step(
        sys: &Self::Sys,
        mut mbox: MooreBox<Self::R, Self::VecC, Self::MatC>,
        step_size: f64,
        previous_pt: &mut Option<VecC>,
        previous_speed: &mut Option<VecC>,
    ) -> anyhow::Result<(MooreBox<Self::R, Self::VecC, Self::MatC>, f64, ProjectivePoint<Self::VecTm>)> {
        if mbox.pt.coord.mag() > sys.chart_switch_threshold {
            #[cfg(debug_assertions)]
            println!("Changing chart ({} to {})...", mbox.pt.chart, mbox.pt.canonical_chart());
            println!("{}", Ctx::get_prec());
            let chart = mbox.pt.canonical_chart();
            mbox = GToChart::to_chart(sys, mbox, chart, 0.75)?;
            mbox.mut_squash();
            *previous_pt = None;
            *previous_speed = None;
        }
        mbox = GRefine::refine(sys, mbox, 0.25)?;
        mbox.mut_squash();
        GExtend::extend(sys, mbox, step_size, previous_pt, previous_speed)
    }
}

pub type RecklessOneStep5 = GOneStep<
    RecklessRefine,
    RecklessExtend5,
    RecklessToChart,
    FixedPrecCtx,
    RecklessProg,
    Rif,
    Cif,
    BoxVec<Cif>,
    BoxMat<Cif>,
    RecklessTm5,
    BoxVec<RecklessTm5>,
>;

pub type RecklessOneStep<const N: usize> = GOneStep<
    RecklessRefine,
    RecklessExtend<N>,
    RecklessToChart,
    FixedPrecCtx,
    RecklessProg,
    Rif,
    Cif,
    BoxVec<Cif>,
    BoxMat<Cif>,
    RecklessTm<N>,
    BoxVec<RecklessTm<N>>,
>;

pub type ArbOneStep<const N: usize> =
    GOneStep<ArbRefine, ArbExtend<N>, ArbToChart, AdaptivePrecCtx, AcbProg, Arb, Acb, AcbVec, AcbMat, AcbTm<N>, AcbVecTm<N>>;

pub struct ArbRecklessOneStep {}

impl OneStep for ArbRecklessOneStep {
    type Sys = AcbRecklessSys;
    type R = Arb;
    type VecC = AcbVec;
    type MatC = AcbMat;
    type Tm = AcbTm<5>;
    type VecTm = AcbVecTm<5>;

    fn one_step(
        sys: &Self::Sys,
        mut mbox: MooreBox<Self::R, Self::VecC, Self::MatC>,
        step_size: f64,
        previous_pt: &mut Option<Self::VecC>,
        previous_speed: &mut Option<Self::VecC>,
    ) -> anyhow::Result<(MooreBox<Self::R, Self::VecC, Self::MatC>, f64, ProjectivePoint<Self::VecTm>)> {
        let current_prec = get_prec();
        if current_prec <= 53 {
            if let Ok(reckless_mbox) = MooreBox::<Rif, BoxVec<Cif>, BoxMat<Cif>>::try_from(&mbox) {
                // dbg!("Conversion sucessful. Going reckless !");
                let mut reckless_prev_pt: Option<BoxVec<Cif>> = match previous_pt {
                    None => None,
                    Some(v) => Some(
                        v.into_iter()
                            .map(|c| c.get_covering_reckless_interval())
                            .collect::<Vec<Cif>>()
                            .into(),
                    ),
                };
                let mut reckless_prev_speed: Option<BoxVec<Cif>> = match previous_speed {
                    None => None,
                    Some(v) => Some(
                        v.into_iter()
                            .map(|c| c.get_covering_reckless_interval())
                            .collect::<Vec<Cif>>()
                            .into(),
                    ),
                };

                match RecklessOneStep5::one_step(
                    &sys.reckless_sys,
                    reckless_mbox,
                    step_size,
                    &mut reckless_prev_pt,
                    &mut reckless_prev_speed,
                ) {
                    Err(_) => return ArbOneStep::one_step(&sys.acb_sys, mbox, step_size, previous_pt, previous_speed),
                    Ok((reckless_mbox, s, _)) => {
                        *previous_pt = match reckless_prev_pt {
                            None => None,
                            Some(v) => Some(v.0.iter().map(|c| c.get_included_acb()).collect::<Vec<Acb>>().into()),
                        };
                        *previous_speed = match reckless_prev_speed {
                            None => None,
                            Some(v) => Some(v.0.iter().map(|c| c.get_included_acb()).collect::<Vec<Acb>>().into()),
                        };
                        mbox = (&reckless_mbox).into();
                        return Ok((
                            mbox,
                            s,
                            ProjectivePoint {
                                coord: AcbVecTm::zeros(1),
                                chart: 0,
                            },
                        ));
                    }
                }
            }
        }
        ArbOneStep::one_step(&sys.acb_sys, mbox, step_size, previous_pt, previous_speed)
    }
}

pub struct GMixed<Sys, F, A> {
    phantom: PhantomData<(Sys, F, A)>,
}

impl<Sys, F, A> OneStep for GMixed<Sys, F, A>
where
    F: OneStep,
    A: OneStep,
    Sys: FaATrait<F = F::Sys, A = A::Sys>,
    FoA<F::R, A::R>: FoATrait<F = F::R, A = A::R>,
    FoA<F::VecC, A::VecC>: FoATrait<F = F::VecC, A = A::VecC>,
    FoA<F::MatC, A::MatC>: FoATrait<F = F::MatC, A = A::MatC>,
    F::VecC: Clone,
    A::VecC: Clone,
    F::MatC: Clone,
{
    type Sys = Sys;
    type R = FoA<F::R, A::R>;
    type VecC = FoA<F::VecC, A::VecC>;
    type MatC = FoA<F::MatC, A::MatC>;
    type Tm = FoA<F::Tm, A::Tm>;
    type VecTm = FoA<F::VecTm, A::VecTm>;

    fn one_step(
        sys: &Self::Sys,
        mut mbox: MooreBox<Self::R, Self::VecC, Self::MatC>,
        step_size: f64,
        previous_pt: &mut Option<Self::VecC>,
        previous_speed: &mut Option<Self::VecC>,
    ) -> anyhow::Result<(MooreBox<Self::R, Self::VecC, Self::MatC>, f64, ProjectivePoint<Self::VecTm>)> {
        let current_prec = get_prec();
        if current_prec <= 53 {
            if let (Ok(fast_coord), Ok(fast_params), Ok(fast_mat)) = (
                mbox.pt.coord.ref_get_fast(),
                mbox.params.ref_get_fast(),
                mbox.mat.ref_get_fast(),
            ) {
                // Clone of the input mb, in the fast arithmetic
                let fast_mbox = MooreBox {
                    pt: ProjectivePoint {
                        coord: fast_coord,
                        chart: mbox.pt.chart,
                    },
                    params: fast_params,
                    mat: fast_mat,
                    rad: mbox.rad,
                };
                let mut fast_prev_pt: Option<F::VecC> = if let Some(FoA::Fast(fast)) = previous_pt {
                    Some(fast.clone()) //TODO: remove clone (this is hard work, &mut Option<T> is annoying)
                } else {
                    None //TODO: maybe try to sloppy convert the vector
                };
                let mut fast_prev_speed: Option<F::VecC> = if let Some(FoA::Fast(fast)) = previous_speed {
                    Some(fast.clone()) //TODO: remove clone (this is hard work, &mut Option<T> is annoying)
                } else {
                    None //TODO: maybe try to sloppy convert the vector
                };
                if let Ok((resmbox, s, tm)) =
                    F::one_step(&sys.get_fast(), fast_mbox, step_size, &mut fast_prev_pt, &mut fast_prev_speed)
                {
                    mbox.pt.coord = FoA::Fast(resmbox.pt.coord);
                    mbox.pt.chart = resmbox.pt.chart;
                    mbox.params = FoA::Fast(resmbox.params);
                    mbox.mat = FoA::Fast(resmbox.mat);
                    mbox.rad = resmbox.rad;
                    *previous_pt = match fast_prev_pt {
                        None => None,
                        Some(v) => Some(FoA::Fast(v)),
                    };
                    *previous_speed = match fast_prev_speed {
                        None => None,
                        Some(v) => Some(FoA::Fast(v)),
                    };
                    return Ok((
                        mbox,
                        s,
                        ProjectivePoint {
                            coord: FoA::Fast(tm.coord),
                            chart: tm.chart,
                        },
                    ));
                }
            }
        }
        let accurate_mbox = MooreBox {
            pt: ProjectivePoint {
                coord: mbox.pt.coord.get_accurate(),
                chart: mbox.pt.chart,
            },
            params: mbox.params.get_accurate(),
            mat: mbox.mat.get_accurate(),
            rad: mbox.rad,
        };
        let mut accurate_prev_pt: Option<A::VecC> = if let Some(FoA::Accurate(accurate)) = previous_pt {
            Some(accurate.clone()) //TODO: remove clone (this is hard work, &mut Option<T> is annoying)
        } else {
            None //TODO: maybe try to sloppy convert the vector
        };
        let mut accurate_prev_speed: Option<A::VecC> = if let Some(FoA::Accurate(accurate)) = previous_speed {
            Some(accurate.clone()) //TODO: remove clone (this is hard work, &mut Option<T> is annoying)
        } else {
            None //TODO: maybe try to sloppy convert the vector
        };
        let (resmbox, ress, restm) = A::one_step(
            &sys.get_accurate(),
            accurate_mbox,
            step_size,
            &mut accurate_prev_pt,
            &mut accurate_prev_speed,
        )
        .unwrap();
        mbox.pt.coord = FoA::Accurate(resmbox.pt.coord);
        mbox.pt.chart = resmbox.pt.chart;
        mbox.params = FoA::Accurate(resmbox.params);
        mbox.mat = FoA::Accurate(resmbox.mat);
        mbox.rad = resmbox.rad;
        *previous_pt = match accurate_prev_pt {
            None => None,
            Some(v) => Some(FoA::Accurate(v)),
        };
        *previous_speed = match accurate_prev_speed {
            None => None,
            Some(v) => Some(FoA::Accurate(v)),
        };
        return Ok((
            mbox,
            ress,
            ProjectivePoint {
                coord: FoA::Accurate(restm.coord),
                chart: restm.chart,
            },
        ));
    }
}

#[cfg(test)]
mod marker_tests {
    use super::{ArbOneStep, ArbRecklessOneStep, GMixed, OneStep, RecklessOneStep, RecklessOneStep5};
    use crate::{foa::FoA, traits::system::AcbRecklessSys};

    fn _marker<T: OneStep>() {}

    #[test]
    fn marker() {
        _marker::<RecklessOneStep5>();
        _marker::<RecklessOneStep<20>>();
        _marker::<ArbOneStep<20>>();
        _marker::<ArbRecklessOneStep>();
        _marker::<GMixed<AcbRecklessSys, RecklessOneStep5, ArbOneStep<5>>>();
    }
}
