use crate::{
    box_vec::BoxVec,
    flint::{Acb, AcbVec, AcbVecTm, Arb},
    reckless::{Cif, Rif},
    reckless_extra::RecklessTm5,
    taylor_model::tm::RecklessTm,
    traits::{vector::Vector, Interval, TaylorModel},
};
use num_traits::Inv;
use std::{
    marker::PhantomData,
    ops::{Add, Mul, Sub},
};

pub trait Predictor {
    type VecC;
    type R;
    type VecTm;

    fn predictor(
        p0: &Self::VecC,
        v0: &Self::VecC,
        p1opt: &Option<Self::VecC>,
        v1opt: &Option<Self::VecC>,
        previous_step_size: f64,
        domain: Self::R,
    ) -> Self::VecTm;
}

pub struct GPredictor<R, C, VecC, VecTm> {
    _r: PhantomData<R>,
    _c: PhantomData<C>,
    _vec_c: PhantomData<VecC>,
    _vec_tm: PhantomData<VecTm>,
}

impl<R, C, VecC, VecTm> Predictor for GPredictor<R, C, VecC, VecTm>
where
    R: From<f64> + Inv<Output = R>,
    C: From<R> + From<i64>,
    VecC: Vector<T = C>
        + Interval
        + for<'a> Mul<&'a C, Output = VecC>
        + Sub<VecC, Output = VecC>
        + for<'a> Add<&'a VecC, Output = VecC>,
    VecTm: TaylorModel<Coeffs = VecC, Domain = R>,
    for<'a, 'b> &'a VecC: Sub<&'b VecC, Output = VecC>,
    for<'a, 'b> &'a VecC: Add<&'b VecC, Output = VecC>,
    for<'a, 'b> &'a VecC: Mul<&'b C, Output = VecC>,
    for<'a, 'b> &'a C: Mul<&'b C, Output = C>,
{
    type R = R;
    type VecC = VecC;
    type VecTm = VecTm;

    #[inline(never)]
    fn predictor(
        p0: &Self::VecC,
        v0: &Self::VecC,
        p1opt: &Option<Self::VecC>,
        v1opt: &Option<Self::VecC>,
        previous_step_size: f64,
        domain: Self::R,
    ) -> Self::VecTm {
        // https://en.wikipedia.org/wiki/Cubic_Hermite_spline
        let h: C = <f64 as Into<Self::R>>::into(previous_step_size)
            .inv()
            .into();
        let mut coeffs = vec![
            p0.clone(),
            v0.clone(),
            Self::VecC::zeros(p0.len()),
            Self::VecC::zeros(p0.len()),
        ];
        if let (Some(_p1), Some(_v1)) = (p1opt, v1opt) {
            let p1 = _p1.squash();
            let v1 = _v1.squash();
            let dp = (p0 - &p1) * &h;
            coeffs[2] = (((v0 * &C::from(2) + &v1) - &dp * &C::from(3)) * &h).squash();
            coeffs[3] = (((v0 + &v1) - &dp * &C::from(2)) * &(&h * &h)).squash();
        }
        VecTm::from_vec(&domain, &coeffs)
    }
}

pub type RecklessPredictor5 = GPredictor<Rif, Cif, BoxVec<Cif>, BoxVec<RecklessTm5>>;

pub type RecklessPredictor<const N: usize> =
    GPredictor<Rif, Cif, BoxVec<Cif>, BoxVec<RecklessTm<N>>>;

pub type ArbPredictor<const N: usize> = GPredictor<Arb, Acb, AcbVec, AcbVecTm<N>>;

#[cfg(test)]
mod marker_tests {
    use super::{ArbPredictor, Predictor, RecklessPredictor, RecklessPredictor5};

    fn _marker<T: Predictor>() {}

    #[test]
    fn marker() {
        _marker::<RecklessPredictor<20>>();
        _marker::<RecklessPredictor5>();
        _marker::<ArbPredictor<10>>();
    }
}
