use super::distort::{ArbDistort, Distort, RecklessDistort};
use crate::{
    box_mat::BoxMat,
    box_vec::BoxVec,
    flint::{AcbMat, AcbVec, Arb, Fmpq},
    precision_ctx::{AdaptivePrecCtx, FixedPrecCtx, PrecCtx},
    prog::{AcbProg, ProgramTrait, RecklessProg},
    projective::{FwBuf, InvDf, MooreBox, ParametricSystem},
    reckless::{Cif, Rif},
    traits::{vector::Vector, Interval},
};
use anyhow::bail;
use nalgebra::Complex;
use num_traits::Inv;
use std::{
    marker::PhantomData,
    ops::{Add, AddAssign, Mul, Sub, SubAssign},
};

pub trait Refine {
    type Sys;
    type R;
    type VecC;
    type MatC;

    fn refine(
        sys: &Self::Sys,
        mbox: MooreBox<Self::R, Self::VecC, Self::MatC>,
        tau: f64,
    ) -> anyhow::Result<MooreBox<Self::R, Self::VecC, Self::MatC>>;
}

pub struct GenRefine<Prog, Ctx, R, VecC, MatC, GDistort> {
    phantom: PhantomData<(Prog, Ctx, R, VecC, MatC, GDistort)>,
}

impl<Prog, Ctx, R, VecC, MatC, GDistort> Refine for GenRefine<Prog, Ctx, R, VecC, MatC, GDistort>
where
    Prog: ProgramTrait<Coeffs = VecC::T>,
    ParametricSystem<Prog>: FwBuf<VecC, S = VecC::T> + InvDf<VecC, MatC, S = VecC::T>,
    Ctx: PrecCtx,
    R: From<f64> + Inv<Output = R> + Interval,
    VecC::T: From<R>,
    for<'r> VecC::T: From<&'r Complex<Fmpq>> + From<&'r R>,
    for<'r> VecC:
        Vector + Add<Output = VecC> + Interval + Mul<&'r VecC::T, Output = VecC> + AddAssign<&'r VecC> + SubAssign<&'r VecC>,
    for<'l, 'r> &'l VecC: Sub<&'r VecC, Output = VecC>,
    MatC: Interval,
    for<'l, 'r> &'l MatC: Mul<&'r VecC, Output = VecC> + Mul<VecC, Output = VecC>,
    GDistort: Distort<VecC, MatC, Sys = ParametricSystem<Prog>, S = VecC::T, MatC = MatC>,
{
    type Sys = ParametricSystem<Prog>;
    type R = R;
    type MatC = MatC;
    type VecC = VecC;

    #[inline(never)]
    fn refine(
        sys: &Self::Sys,
        mut mbox: MooreBox<Self::R, Self::VecC, Self::MatC>,
        tau: f64,
    ) -> anyhow::Result<MooreBox<Self::R, Self::VecC, Self::MatC>> {
        let mut coord;
        let mut inv_df;
        let mut f;

        let param = VecC::T::from(&mbox.params);

        // Checks on input in debug mode
        #[cfg(debug_assertions)]
        {
            inv_df = sys.inv_df(&param, &mbox.pt)?;
            f = sys.f(&param, &mbox.pt);
            let krawczyk = (&inv_df * &f) * &VecC::T::from(R::from(mbox.rad).inv())
                + GDistort::distort(sys, &param, &mbox.pt, &inv_df, mbox.rad);
            if krawczyk.mag() >= 1.0 {
                println!(
                    "Warning: the input box does does not pass the criterion in the arithmetic (||𝐊|| = {})",
                    krawczyk.mag()
                );
            }
        }

        mbox.mut_squash(); // This is for safety, else termination is not ensured. This keeps the property of mbox, by def.
        let mut new_rad = mbox.rad * 1.1;

        // We first try to perform 2 quasi-newton iterations.
        for _ in 0..2 {
            let delta = &mbox.mat * sys.f(&param, &mbox.pt);
            coord = &mbox.pt.coord - &delta;
            // For correction, one must have 20⋅diam(x - δ) ≤ ||δ||, so we do this check rigorously
            //TODO: maybe make diam and mag return Rifs
            if !((Rif::from(coord.diam()) * Rif::from(20.0))._sup() <= delta.mag_lower()) {
                break;
            }
            coord.mut_squash();
            mbox.pt.coord = coord;
        }

        while {
            // Krawczyk test. Quick test first, then full test. If it passes, mbox is a τ-mb
            inv_df = sys.inv_df(&param, &mbox.pt)?;
            inv_df.mut_squash(); // Unchecked arithmetic.
            f = sys.f(&param, &mbox.pt);
            let krawczyk_lhs = (&inv_df * &f) * &VecC::T::from(R::from(new_rad).inv()); // A⋅□f(x)/{r}
            krawczyk_lhs.mag() * 7.0 > tau
                || (krawczyk_lhs + GDistort::distort(sys, &param, &mbox.pt, &inv_df, new_rad)).mag() > tau
        } {
            // the test has failed, we must improve something
            let delta = &mbox.mat * &f;
            if delta.mag() * 64.0 <= new_rad * tau {
                new_rad *= 0.5;
                if (new_rad < 1e-20 || new_rad * 32.0 <= mbox.rad * tau) && Ctx::set_prec(Ctx::get_prec() + 2).is_err() {
                    bail!("new radius is too small");
                }
            } else if {
                coord = &mbox.pt.coord - &delta;
                (Rif::from(coord.diam()) * Rif::from(20.0))._sup() <= delta.mag_lower()
            } {
                coord.mut_squash();
                mbox.pt.coord = coord;
            } else if Ctx::set_prec(Ctx::get_prec() + 1).is_err() {
                bail!("the roundoff error in the Newton iteration exceeds 5%");
            }
        }
        mbox.rad = new_rad;
        mbox.mat = inv_df;

        //TODO: analyze this more in details
        // Finally, we try growing the radius.
        // let f = sys.f(&mbox.params, &mbox.pt);
        // while 2.0 * mbox.rad < 1.0 && ((&mbox.mat * &f) * &VecC::T::from(R::from(2.0 * mbox.rad).inv())
        //     + GDistort::distort(sys, &mbox.params, &mbox.pt, &mbox.mat, 2.0 * mbox.rad))
        // .mag()
        //     <= tau
        // {
        //     mbox.rad *= 2.0;
        // }

        Ok(mbox)
    }
}

pub type RecklessRefine = GenRefine<RecklessProg, FixedPrecCtx, Rif, BoxVec<Cif>, BoxMat<Cif>, RecklessDistort>;

pub type ArbRefine = GenRefine<AcbProg, AdaptivePrecCtx, Arb, AcbVec, AcbMat, ArbDistort>;

#[cfg(test)]
mod marker_tests {
    use super::{ArbRefine, RecklessRefine, Refine};

    fn _marker<T: Refine>() {}

    #[test]
    fn marker() {
        _marker::<RecklessRefine>();
        _marker::<ArbRefine>();
    }
}
