use super::distort::{ArbDistort, Distort, RecklessDistort};
use crate::{
    box_mat::BoxMat,
    box_vec::BoxVec,
    flint::{AcbMat, AcbVec, Arb, Fmpq},
    precision_ctx::{AdaptivePrecCtx, FixedPrecCtx, PrecCtx},
    prog::{AcbProg, FlintProg, ProgramTrait, RecklessProg},
    projective::{FwBuf, InvDf, MooreBox, ParametricSystem, ProjectivePoint},
    reckless::{Cif, Rif},
    traits::{vector::Vector, Interval},
};
use anyhow::bail;
use nalgebra::Complex;
use num_traits::{Inv, One, Zero};
use std::{
    fmt::Debug,
    marker::PhantomData,
    ops::{Add, AddAssign, Div, Mul, Sub, SubAssign},
};

pub trait ToChart {
    type Sys;
    type R;
    type VecC;
    type MatC;

    fn to_chart(
        sys: &Self::Sys,
        mbox: MooreBox<Self::R, Self::VecC, Self::MatC>,
        chart: usize,
        tau: f64,
    ) -> anyhow::Result<MooreBox<Self::R, Self::VecC, Self::MatC>>;
}

pub struct GenToChart<Prog, Ctx, R, VecC, MatC, GDistort> {
    phantom: PhantomData<(Prog, Ctx, R, VecC, MatC, GDistort)>,
}

impl<Prog, Ctx, R, VecC, MatC, GDistort> ToChart for GenToChart<Prog, Ctx, R, VecC, MatC, GDistort>
where
    Prog: ProgramTrait<Coeffs = VecC::T>,
    Ctx: PrecCtx,
    R: From<f64> + Inv<Output = R> + Interval,
    for<'r> VecC::T: From<R>
        + From<&'r R>
        + Interval
        + Zero
        + One
        + From<&'r Complex<Fmpq>>
        + Div<&'r VecC::T, Output = VecC::T>
        + From<&'r Complex<Fmpq>>,
    for<'r> VecC: Debug
        + Vector
        + Add<Output = VecC>
        + Interval
        + Mul<&'r VecC::T, Output = VecC>
        + AddAssign<&'r VecC>
        + SubAssign<&'r VecC>
        + Sub<&'r VecC, Output = VecC>
        + Add<&'r VecC, Output = VecC>,
    MatC: Interval,
    ParametricSystem<Prog>: FwBuf<VecC, S = VecC::T> + InvDf<VecC, MatC, S = VecC::T>,
    GDistort: Distort<VecC, MatC, Sys = ParametricSystem<Prog>, S = VecC::T, MatC = MatC>,
    for<'l, 'r> &'l MatC: Mul<&'r VecC, Output = VecC> + Mul<VecC, Output = VecC>,
    for<'l, 'r> &'l VecC: Sub<&'r VecC, Output = VecC>,
{
    type Sys = ParametricSystem<Prog>;
    type R = R;
    type MatC = MatC;
    type VecC = VecC;

    #[inline(never)]
    fn to_chart(
        sys: &Self::Sys,
        mut mbox: MooreBox<Self::R, Self::VecC, Self::MatC>,
        chart: usize,
        tau: f64,
    ) -> anyhow::Result<MooreBox<Self::R, Self::VecC, Self::MatC>> {
        mbox.mut_squash(); // Safety measure
        let mut new_rad = mbox.rad * 1.1;
        let mut new_pt_coord;
        let mut mat_in_chart;
        let mut rad_in_chart;
        let mut pt_in_chart;
        let param = VecC::T::from(&mbox.params);

        while {
            // Compute center and radius of the candidate box.
            pt_in_chart = mbox.pt.to_chart(chart);
            pt_in_chart.coord.mut_squash();
            //TODO: this a bit annoying to write
            rad_in_chart = (ProjectivePoint {
                coord: VecC::from_value(mbox.pt.coord.len(), VecC::T::square(new_rad)) + &mbox.pt.coord,
                chart: mbox.pt.chart,
            }
            .to_chart(chart)
            .coord
                - &pt_in_chart.coord)
                .mag();

            mat_in_chart = sys.inv_df(&param, &pt_in_chart)?;
            let krawczyk_lhs = (&mat_in_chart * sys.f(&param, &pt_in_chart)) * &VecC::T::from(R::from(rad_in_chart).inv());
            krawczyk_lhs.mag() * 7.0 > tau
                || (GDistort::distort(sys, &param, &pt_in_chart, &mat_in_chart, rad_in_chart) + &krawczyk_lhs).mag() > tau
        } {
            // the test has failed, we must improve something
            let delta = &mbox.mat * &sys.f(&param, &mbox.pt);
            if delta.mag() * 64.0 <= new_rad * tau {
                new_rad *= 0.5;
                if (new_rad < 1e-20 || new_rad * 32.0 <= mbox.rad * tau) && Ctx::set_prec(Ctx::get_prec() + 2).is_err() {
                    bail!("new radius is too small");
                }
            } else if {
                new_pt_coord = &mbox.pt.coord - &delta;
                (Rif::from(new_pt_coord.diam()) * Rif::from(20.0))._sup() <= delta.mag_lower()
            } {
                new_pt_coord.mut_squash();
                mbox.pt.coord = new_pt_coord;
            } else if Ctx::set_prec(Ctx::get_prec() + 1).is_err() {
                bail!("the roundoff error in the Newton iteration exceeds 5%");
            }
        }

        mbox.rad = rad_in_chart;
        mbox.pt = pt_in_chart;
        mbox.mat = mat_in_chart;
        Ok(mbox)
    }
}

pub type RecklessToChart = GenToChart<RecklessProg, FixedPrecCtx, Rif, BoxVec<Cif>, BoxMat<Cif>, RecklessDistort>;

pub type ArbToChart = GenToChart<AcbProg, AdaptivePrecCtx, Arb, AcbVec, AcbMat, ArbDistort>;

#[cfg(test)]
mod marker_tests {
    use super::{ArbToChart, RecklessToChart, ToChart};

    fn _marker<T: ToChart>() {}

    #[test]
    fn marker() {
        _marker::<RecklessToChart>();
        _marker::<ArbToChart>();
    }
}
