use algpath::{
    flint::{Arb, Fmpq},
    precision_ctx::{get_prec, set_prec},
    traits::{
        interval::{IntervalComparisons, RealInterval},
        Interval,
    },
};
use flint_bindings::*;
use rand::Rng;
use std::mem::MaybeUninit;

#[test]
fn ord() {
    let mut rng = rand::rng();
    for _ in 0..1000 {
        let mut a: f64 = rng.random();
        let mut b: f64 = rng.random();
        if b <= a {
            (a, b) = (b, a);
        }
        let x = Arb::from_inf_sup(a.into(), b.into());

        let mut c: f64 = rng.random();
        let mut d: f64 = rng.random();
        if d <= c {
            (c, d) = (d, c);
        }
        let y = Arb::from_inf_sup(c.into(), d.into());

        if b <= c {
            assert!(x.interval_le(&y), "le fail");
        } else if d <= a {
            assert!(x.interval_ge(&y), "ge fail");
        } else {
            assert!(!x.interval_ge(&y) && !x.interval_ge(&y), "overlap fail");
        }
    }
}

#[test]
fn square() {
    let x = Arb::square(1.0);
    println!("{:?}", x); // TODO check strange behavior on mag (which is a bit more than 1)
}

#[test]
fn squash() {
    let mut x = Arb::from_inf_sup(0.0.into(), 0.5.into());
    let y = x.squash();
    println!("x: {:?}", x);
    println!("x.squash(): {:?}", y);
    x.mut_squash();
    println!("x after mut_squash(): {:?}", x);
    assert!(x == y);
}

#[test]
fn mag() {
    let mut x = Arb::from_inf_sup(0.0.into(), 1.0.into());
    println!("{}", x.mag());
    assert!(x.mag() >= 1.0);
    x = Arb::square(2.0);
    assert!(x.mag() >= 2.0);
    println!("{}", x.mag());
}

#[test]
fn mag_rand() {
    set_prec(200);
    let mut rand_gen = MaybeUninit::uninit();
    unsafe {
        let mut x: Arb = Arb::default();
        flint_randinit(rand_gen.as_mut_ptr());
        let mut rand_gen = rand_gen.assume_init();
        for _ in 0..1000 {
            arb_randtest(&mut x.0, &mut rand_gen, get_prec(), 30);
            println!("{}", x.mag());
            println!("{}", x.mag_lower());
            assert!(x.mag_lower() <= x.mag());
            assert!(0.0 <= x.mag_lower());
        }
        flint_randclear(&mut rand_gen);
    }
}

#[test]
fn diam() {
    let x = Arb::square(8.0);
    assert!(x.diam() >= 16.0);
}

#[test]
fn from_f64() {
    let _: Arb = 2.0.into();
}

#[test]
fn intersection() {
    let x = Arb::entire();
    let y = Arb::from_inf_sup(0.0.into(), 1.0.into());
    println!("{}", y.intersection(&x));
}

#[test]
fn get_covering_reckless_interval() {
    set_prec(100);
    let x = Fmpq::from_machine_int(1, 3);
    let mut box_x: Arb = x.into();
    box_x.mut_squash();
    dbg!(&box_x);
    dbg!(box_x.get_covering_reckless_interval());
}

#[test]
fn try_get_included_reckless_interval() {
    let mut x: Arb = 1.2.into();
    assert!(x.try_get_included_reckless_interval().is_ok());
    set_prec(100);
    x = Fmpq::from_machine_int(1, 3).into();
    let y: Arb = Fmpq::from_machine_int(1, 3).into();
    assert!(x.try_get_included_reckless_interval().is_err());
    assert!(y.try_get_included_reckless_interval().is_err());
    set_prec(53);
    assert!((x * y).try_get_included_reckless_interval().is_ok());
}

#[test]
fn display() {
    set_prec(500);
    let x = Fmpq::from_machine_int(1, 3);
    let mut x = Arb::from(x);
    println!("{x}");
    x = Arb::from(1.1);
    println!("{x}");
}
